/*
 * Copyright (C) Microsoft Corporation, 1982
 */

/*      Modification History
 *
 *      05/17/82        JGL     M001
 *              - modify getx() into in-line GETX(), ala putc in stdio
 *                This produces a significant performance gain.
 */

#include <stdio.h>

#define MAXLINE 128
#define MAXINSRT 10

/* symbol table definitions */
#define MAXSYM 60
#define HASHVAL 419
#define A_TEXT	2
#define A_DATA	4
#define A_BSS	6
#define A_ABS	0
#define A_UEXT	8
#define A_EXPR	1
#define A_UND	3
#define F_EXTRN 1
#define F_FIXED 2
#define F_BUSY  4
#define F_LOCAL 8

/* magic numbers for instruction decoding */

#define BYTE 1
#define WORD 2

/* arithmetic operations */

#define I_ADD	0
#define I_ADC	2
#define I_AND	4
#define I_CMP	7
#define I_OR	1
#define I_SUB	5
#define I_SBB	3
#define I_TEST	8
#define I_XOR	6

/* control transfer within segment */
#define I_CALL 1
#define I_JUMP 2

#define I_DEC 0
#define I_INC 1


#define AX_AL 0
#define CX_CL 1


struct optab {
	char *op_text;
	int op_token;
	int op_typ;
	int op_styp;
} op_codes[];

/* the expression stack stuff */

#define E_OFFS 0x100
#define E_BASE 0x200
#define E_SELF 0x400

struct estak {
	int e_type; /* type of operand ( same as symbtab types ) */
	union {
		int e_int;  /* integer value of expression */
		struct Expr *e_expr;
		struct Sym *e_sym;
	} e_v;
};
struct Expr {
	int E_opr; /* expression operator */
	struct estak E_a,E_b; /* operands */
};

/* memory operand stack temporaries */
struct memstk {
	int m_mod;
	struct estak *m_disp;
} ;


struct Sym {
	int sy_type;
	int sy_snum;
	unsigned sy_value;
	struct Sym *sy_link;
	struct Sym * sy_glink;
	char sy_flags;
	union {				/* note: this must be the last 
				   	   item in this structure */
		char sy_text[2];
		int sy_int;
	} sy_nam;
} ;

#define sy_name sy_nam.sy_text

extern struct Sym Dot ;
extern struct Sym *lookup();
extern struct Sym *gl_head,*gl_tail;	/* global symbol chain */
extern struct Sym *hashtab[];
extern struct Sym *lcl_tab;		/* local symbols chain */


#define etyp(x) (x)->e_type
#define eint(x) ((x)->e_v.e_int)
#define esym(x) ((x)->e_v.e_sym)

#define optyp(x) (x)->op_typ
#define opstyp(x) (x)->op_styp
#define optoken(x) (x)->op_token

char *mktemp();
char *malloc();

char indxbit[];		/* these two arrays are used to compute the */
char r_m[] ;		/* r/m field for indexed memory references */

struct estak *c_value;	/* expression index for value of constant and */
char c_len;		/* length ( '*' = 8, '#' = 16 ) */

struct estak estk[];	/* expression stack temporaries */
struct estak *e,*exp;

struct memstk MMem[];	/* memory operand stack temporaries */
struct memstk *m,*mem;
struct memstk *mem_com();

#define GETX(x)  (readflg?getx():(*(lineptr++)))        /* M001 */

int word,ilen;		/* miscellaneous variables & temporaries */
int reg,i;
unsigned Text ;		/* size of text segment */
unsigned Data ;		/* size of data segment */
unsigned Bss ;		/* size of bss segment */
unsigned Entry ;		/* start location for program */

#define dot Dot.sy_value	/* location counter at start of this line */
unsigned new_dot ;	/* updated location counter */
int Linenum ;		/* current line being assembled */
int Globals ;		/* index number for output symbol table */
char Pass;		/* current pass of assembler */

				/* array to save generated code for listing */
int j ;			/* index into core array */
char core[];
				/* flags in relocation data for temp file */
#define R_BYT 020
#define R_PC_REL 040
#define R_LINE 128
#define R_TEXT 129
#define R_DATA 130
#define R_BSS 131
#define R_END 132

/* #if RELFMTI no need to conditional these out */
#define R_TARG	133		/* change frame to target */
#define R_SELFR 134		/* self relative fixup */
#define R_SEGR  135		/* segment relative fixup */
#define R_FDATA 136		/* change frame to data for next reference */
		/* note these must be last four */
#define R_FTEXT 137		/* change frame to text for next reference */
#define R_POFFS 138		/* long pointer offset relocatable */
#define R_PBASE 139		/* long pointer base relocatable */
#define R_LNPTR	140		/* long pointer relocatable */
/* #endif */


FILE *list,*input;	/* file pointers for source input and listing */
FILE *inpstk[];		/* stack of insert files */
int istak;		/* index into input stack */

FILE *object;		/* temporary work file, pseudo object code */
char *objtemp ;
				/* flags for lexical analyzer and line buffer */
char opcflg;		/* 1 = still looking for opcode on this line */
char readflg;		/* 1 = need to read another line */
char operflg ;		/* 1 = octal & hex constants are valid */
char endflag ;		/* 1 = found end statement */
char linebuf[];
char *lineptr;

char String[];		/* buffer for accumlating delimited strings*/
char *strptr;		/* current character of string */
char delimit;		/* string delimiter character */
char strmod;		/* 1= string scan mode */

char *reg_nam[];
int reg_tok[];
char reg_num[];
char reg_len[];

int arop_ai[];
int arop_rm[];

char Listing ;		/* 1 = produce listing on first pass */
char Ext_only ;		/* 1 = put only .globals in text file */
char Frm_cmp ;		/* 1 = compiler generated assembly */
char Und_glb ;		/* 1 = make undefined syms into globals */
char Med_opt ;		/* 1 = medium model */
int Errors ;		/* number of errors */
int no_eof ;		/* no end of file */
char *last_glb;

/* #if RELFMTI no need to comment this out */
char *Rel_name;		/* relocation identifier name */
char Rel_lng;		/* long pointer relocation flag */
char Cur_Frame;		/* current relocation frame */
char out_lines;		/* output line records */
char end_out;		/* _end definition output */
char Self_rel;		/* self relative flag */
char Fix_it;		/* fixed location symbol */
/* #endif */

#define WAIT8087 0xcd		/* emulator os interrupt */
