/*
 * Copyright (C) Microsoft Corporation, 1982
 */

#ifdef IEEE
#define LOWBITS	 0x0007			/* low bits to be shifted */
#define DBLEXP	 0x7ff0			/* double precision exponent position */
#define HSTEXP	 0x7f80			/* host exponent position */
#define SIGEXP	 0x7f80			/* single precision exponent position */
#define EXPDIF	 3			/* diff in host,target exponent size */
#define VALBITS	 13			/* valid bits left in word */
#define SIGNBIT	 0x8000			/* sign bit */

static union
	{
	double f;
	unsigned int p[4];
	} cn ;

hdtots(val)
double val;
	{
	int sign,i,exp;
	cn.f = val;
#ifdef FLTDBG
	printf("hdtots of %f\n",val) ;
#endif
	if(val == 0.0)
		return(0.0) ;
	exp = (cn.p[0] & HSTEXP) >> 7;
	exp = (exp-2)<<7;
					/* exponent -1 for diff in binary
					   assumption,-1 for diff in bias */
	sign = cn.p[0] & SIGNBIT ;	/* sign */
#ifdef FLTDBG
	printf("%x,%x,%x\n",cn.p[0] & ~SIGEXP,exp,sign) ;
#endif
	cn.p[0] = (cn.p[0] & ~SIGEXP) | exp | sign ;	
					/* store new exponent */
#ifdef FLTDBG
	for(i=0;i<4;i++)
		printf("%x ",cn.p[i]) ;
	printf("\n") ;
#endif
	return(cn.f) ;
	}

hdtotd(val)
double val;
	{
	int i,j,exp,movbits,sign;
	cn.f = val;
#ifdef FLTDBG
	printf("hdtotd of %f\n",val) ;
	for(i=0;i<4;i++)
		printf("%x ",cn.p[i]) ;
	printf("\n") ;
#endif
	if(val == 0.0)
		return(0.0) ;		/* test for zero */
	exp = (cn.p[0] & HSTEXP) >> 7 ;
	exp = (exp+894) << 4 ;
					/* exponent -1 for diff in binary
					   assumption,-1 for diff in bias
					   -128 for host bias 
					   +1023 for target bias */
#ifdef FLTDBG
	if(cn.p[3] & LOWBITS)
		printf("bit of precision in fraction lost\n") ;
#endif
	for(i=2,j=3;j>0;i--,j--)
		{
		movbits = cn.p[i] & LOWBITS ;
					/* get bits from previous words */
		cn.p[j] = (cn.p[j] >> EXPDIF) | (movbits << VALBITS) ;
					/* shift into current word */
		}
	sign = cn.p[0] & SIGNBIT ;	/* save the sign bit */
	cn.p[0] >>=EXPDIF ;		/* make room for larger exponent */
#ifdef FLTDBG
	printf("%x,%x,%x\n",cn.p[0] & ~DBLEXP,exp,sign) ;
#endif
	cn.p[0] = (cn.p[0] & ~DBLEXP) | sign | exp;
#ifdef FLTDBG
	for(i=0;i<4;i++)
		printf("%x ",cn.p[i]) ;
	printf("\n") ;
#endif
	return(cn.f) ;
	}
#else
	no conversion routines coded
#endif
