/*
 * Copyright (C) Microsoft Corporation, 1982
 */

#include "y.tab.h"
#define NOOPCDEF 1
#include "a86.h"
#include <time.h>
#include <ctype.h>
#ifdef HEADER
#include "a.out.h"
#endif

#define loc (*locptr)
static unsigned *locptr = &Text;

#define NAMEREC 0x96
#define TMODREC 0x80
#define SEGDEF  0x98
#define GRPDEF  0x9A
#define FIXUP	0x9c
#define DATAREC	0xa0
#define EXTNAMS	0x8c
#define LINEREC 0x94
#define MODEND  0x8a
#define COMREC	0x88
#define TYPDEF	0x8e

#define TEXTN	1			/* text name index */
#define DATAN	2			/* data name index */
#define BSSN	3			/* bss name index */
#define GROUPN	4			/* group name index */
#define EODN	9			/* end of data,_end segment name */

#define TEXTS	1			/* text segment index */
#define DATAS	2			/* data segment index */
#define BSSS	3			/* bss segment index */
#define EODS	4			/* end segment */

#define GROUPG	1			/* dgroup group index */

#define TTXTTHRD 0			/* target text thread */
#define TDATTHRD 1			/* target in data thread */
#define TBSSTHRD 2			/* target in bss thread */
#define FTXTTHRD 0			/* frame from text thread */
#define FDGPTHRD 0x10			/* frame from group thread (1<<4) */
#define F_EXT 	 0x8		/* target not specified by thread, external */

#define NAMECHRS 31 	/* _TEXT _DATA _BSS DGROUP CODE DATA BSS */
#define MAX_REC 1022	/* leave room for long pointers */
#define MAXCOMMON 0xffffffL
#define DATAHD	6	/* size of data record header */

#define Putc(c) *RECPTR = (c); RECPTR++; RECCNT++
#define Putw(w) RECTMP = w;\
		RECCP = (char *)&RECTMP;\
		*RECPTR++ = *RECCP++;\
		*RECPTR++ = *RECCP; RECCNT += 2

#define Putrw(w) RECTMP = w;\
		RECCP = (char *)&RECTMP;\
		*RECPTR++ = *(RECCP+1);\
		*RECPTR++ = *RECCP; RECCNT += 2
static char RECORD[MAX_REC+4],*RECPTR = &RECORD,*RECCP,*RECSTART = &RECORD ;
static int  RECTMP,*RECWP,RECCNT = 0 ;		/* record formatting buffer */
static long RECLTMP;

#define LINEMAX 200				/* line number recording */
static int Lines[LINEMAX] = {1,0},Linecnt = 2;
static int SETLOCF = 1 ;

static FILE *output;

#ifdef HEADER
static struct xexec header;
#endif

#define FIXREC 125
#define FIX fixup[FIXCNT]
struct fixup
	{
	int	loc_addr;
	char	loc_pos ;
	char	frame ;
	char	target;
	int	index;
	} fixup[FIXREC] ;
static int FIXCNT = 0;
static int lngout = 0;
static int lastloc[A_BSS+1] ;

Puts(s) 
char *s ;
	{
	while(*(s)) 
		{
		*RECPTR++ = *s++;
		RECCNT++;
		}
	}

Check() 
	{
	register int check,i ;
	i=0 ;
	for(check=0,RECCP=RECSTART;RECCP != RECPTR;i++,RECCP++)
		{
#if DEBUG
		dbprt("chk","%02.2x  ",*RECCP) ;
		if(i == 12)
			{
			i = 0 ;
			dbprt("chk","\n") ;
			}
#endif
		check += *RECCP;
		}
#if DEBUG
	dbprt("chk","checksum value %x,",check) ;
#endif
	*RECCP = -(check%256);		/* set the checksum */
	RECCNT++ ;
#if DEBUG
	dbprt("chk","checksum %02.2x\nrecord length %d\n",*RECCP,RECCNT) ;
#endif
	RECPTR = RECSTART;
	fwrite(RECPTR,sizeof(char),RECCNT,output) ;
	RECCNT = 0 ;
	}

dumprel()
{
	register int i;
	int val ;
	int savepos,count;
#if DEBUG
	dbprt("rel","dumprel called,FIXCNT = %d\n",FIXCNT) ;
#endif
	if(FIXCNT == 0)
		return;
	Putc(FIXUP) ;
	val = 0 ;			/* start with zero length */
	Putw(val) ;
	for(i=0;i < FIXCNT;i++)
		{			/* loop through fixup records */
		val = (fixup[i].loc_pos << 10) | fixup[i].loc_addr;
		Putrw(val) ;		/* LOCAT field reversed byte */
		if(fixup[i].target != F_EXT)
			{		/* thread references */
			val = 0x0c | fixup[i].frame | fixup[i].target;
			Putc(val) ;
			}
		else
			{		/* external reference */
			val = 0x06 | fixup[i].frame ; 
			Putc(val) ;
			if(fixup[i].index < 128)
				{	/* one byte index */
				Putc(fixup[i].index) ;
				}
			else
				{	/* word index */
				val = fixup[i].index | 0x8000;
				Putw(val) ;
				} ;
			}
		}
	count = RECCNT ;
	savepos = RECPTR ;
	RECPTR = RECSTART+1 ;		/* position to length field */
	Putw(count-2) ;
	RECCNT = count ;
	RECPTR = savepos ;
	Check() ;			/* checksum and output */
	FIXCNT = 0 ;			/* clear fixup array */
} ;

dumpbuf(typ,maxloc)
int typ,maxloc ;
{
	register int count ;
	register char *savepos;
	int length ;
	count = RECCNT ;
	savepos = RECPTR ;		/* save record position */
	RECPTR = RECSTART+1 ;
	length = count-2 ;		/* total record length */
	Putw(length) ;
	RECCNT = count ;		/* restore record count */
	RECPTR = savepos;
	Check() ;			/* checksum and output record */
#ifdef DEBUG
	dbprt("rel","dumpbuf length %d,maxloc %x,type %d\n",length,maxloc,typ);
#endif
	dumprel() ;			/* output any fixups */
	SETLOCF = 1 ;			/* new data record flag */
	lastloc[typ] = maxloc ;		/* next starting loc for this typ */
} ;

fillbuf(siz,val,loctyp,newloc)
int siz,val,loctyp,newloc;
{
int length = 0 ;
	if(loctyp == A_BSS)
		return;			/* no bss output */
	if ((RECCNT+siz) > MAX_REC)
		{			/* no overflow,but dont split long */
#ifdef DEBUG
		dbprt("rel","fillbuf calling dumpbuf to empty buffer\n") ;
#endif
		dumpbuf(loctyp,newloc-siz) ;/* dump the buffer */
		}
	if(SETLOCF)
		{			/* ouput starting location */
		Putc(DATAREC) ;		/* data record */
		Putw(length) ;		/* start with 0 length */
		if (loctyp == A_TEXT)
			{
			Putc(TEXTS) ;	/* segment index for text */
			}
		else
			{
			Putc(DATAS) ;	/* segment index for data */
			}
		Putw(lastloc[loctyp]) ;
		SETLOCF = 0 ;		/* clear the flag */
#ifdef DEBUG
		dbprt("rel","fillbuf initializing record,loc %x,type %d\n",
		lastloc[loctyp],loctyp) ;
#endif
		}
	if(siz == 1)
		{
		Putc(val) ;		/* output the byte */
		}
	else
		{
		Putw(val) ;		/* output the word */
		if(lngout)
			{
			Putw(0) ;	/* 2nd word of long ptr */
			getw(object) ;	/* skip over padding - typ,byte */
			getw(object) ;	/* there are two of them */
			loc += 2;	/* count the 2nd word */
			}
		}
lngout = 0 ;				/* always clear flag to be safe */
}

dumpline(loctype)
int loctype ;
	{
	int length ;
	register int check ;
	register char *cp ;
	if (!Linecnt)
		return ;
	if (out_lines)
		{
		putc(LINEREC,output) ;
		check = LINEREC ;
		length = (Linecnt*2) + 3 ;
		cp = (char *) &length ;
		putc(*cp,output) ;
		check += *cp++;
		putc(*cp,output) ;
		check += *cp ;
		if(loctype == A_TEXT)
			{
			putc(0,output) ;	/* no group */
			putc(TEXTS,output) ;		/* text index */
			check += TEXTS ;
			}
		else
			{
			putc(GROUPG,output) ;
			check += GROUPG ;
			if (loctype == A_DATA)
				{
				putc(DATAS,output) ;
				check += DATAS ;
				}
			else
				{
				putc(BSSS,output) ;
				check += BSSS ;
				}
			}
		for(i=0;i<Linecnt;i++)
			{
			cp = &Lines[i] ;
			putc(*cp,output) ;
			check += *cp++;
			putc(*cp,output) ;
			check += *cp ;
			cp = &Lines[++i] ;
			putc(*cp,output) ;
			check += *cp++;
			putc(*cp,output) ;
			check += *cp ;
			}
		check %= 256 ;
		putc(check,output) ;
		}
	Linecnt = 0 ;
	}

int switchloc(Plocptr,Ploctype,Pnewloc,type)
int **Plocptr,*Pnewloc,*Ploctype,type ;
	{
	if(RECCNT >DATAHD)		/* dump previous segment */
		{
#ifdef DEBUG
		dbprt("rel","switchloc calling dumpbuf buf typ %d,max loc %x\n",
		*Ploctype,**Plocptr) ;
#endif
					/* save max offset for this segment*/
		dumpbuf(*Ploctype,**Plocptr) ;
					/* dump buffer and initialize
					   for new location offset */
		}
	Linecnt -= 2 ;	/* remove this line */
	dumpline(*Ploctype) ;		/* dump line records */
	*Plocptr = Pnewloc;		/* switch location counters */
	*Ploctype = type;		/* extablish new type */
	Lines[0] = Linenum ;		/* initialize line numbers */
	Linecnt = 2;
	return(Lines[1] = **Plocptr) ;
	}

static struct Sym *Sym;			/* for UEXT symbol table pointers */

fix_up(name)
char *name;
{
	register int val,rel;
	register int typ;
	int length,nc;
	int relsize = 0,mval,mrel ;
	int self_rel = 0;			/* self relative flag */
	char *cp ;
	struct Expr *expr;
	int old_loc = 0;
	int reltyp = 0;			/* default to segment relative */
	int loctype = A_TEXT;
	int ovnindx = 8 ;
	int textfram = 0 ;		/* change frame of reference flag */
	int targfram = 0 ;		/* change frame of reference flag */
	long ftell() ;
#ifdef DEBUG
	dbprt("trc","trc - fixup called\n") ;
#endif

					/* initialize for output */
	output = fopen(name,"w");
	if (output == NULL)
		{			/* output file error */
		fprintf(stderr,"cannot open output file %s\n",name);
		exit(4);
		}
#ifdef HEADER
	header.x_magic = X_MAGIC;
	header.x_text = Text;
	header.x_data = Data;
	header.x_bss  = Bss;
	header.x_cpu  = XC_8086;
	header.x_relsym = XR_R86REL | XR_S86REL;
	header.x_renv = XE_V2;
	if(Med_opt)
		header.x_renv |= XE_LTEXT ;
	fwrite((char *)&header,sizeof(struct xexec),1,output) ;
#endif
	for(cp = Rel_name;*cp;cp++)
		if(islower(*cp)) *cp = *cp - 0x20;
#if DEBUG
	dbprt("trc","trc - fix_up start of record output\n") ;
#endif
					/* build T-MODULE RECORD */
	Putc(TMODREC);			/* name the translation */
	nc = strlen(Rel_name);
	length = nc + 2;
	Putw(length) ;			/* length of record */
	Putc(nc) ;			/* length of name */
	Puts(Rel_name) ;		/* output name */
#if DEBUG
	dbprt("chk","chk - T-MODULE RECORD\n") ;
#endif
	Check() ;				/* evaluate the checksum */
					
					/* build COMMENT RECORD */
	Putc(COMREC) ;
	length = 26 ;
	Putw(length) ;			/* record size */
	Putw(0) ;
	Puts("XENIX 8086 ASSEMBLER - ") ;
	Check() ;
					/* build general NAMES RECORD */
	if(!Med_opt)
		nc = 0;			/* small model, one text segment */
	Putc(NAMEREC) ;
	if(end_out)
		{
		ovnindx = 10 ;
		length=nc+NAMECHRS+18;	/* Rel_name prefix,names,11 no.chars */
		}
	else				/* EOD _EOD(7) and checksum	     */
		length = nc+NAMECHRS+9;	/* Rel_name prefix,names,8 no.chars */
	Putw(length) ;			/* and checksum is length of record */
	Putc(nc+5) ;
	if(Med_opt)
		Puts(Rel_name) ;
	Puts("_TEXT") ;			/* name 1 is [Rel_name]_TEXT */
	Putc(5) ;
	Puts("_DATA") ;			/* name 2 is DATA */
	Putc(4) ;
	Puts("_BSS") ;			/* name 3 is BSS */
	Putc(6) ;
	Puts("DGROUP") ;		/* name 4 is DGROUP */
	Putc(4) ;
	Puts("CODE") ;			/* name 5 is CODE */
	Putc(4) ;
	Puts("DATA") ;			/* name 6 is DATA */
	Putc(3) ;
	Puts("BSS") ;			/* name 7 is BSS */
	if(end_out)
		{			/* name 8 might be EOD */
		Putc(3) ;
		Puts("EOD") ;
		Putc(4) ;		/* name 9 might be _EOD */
		Puts("_EOD") ;
		}
	Putc(0) ;			/* name 8 is empty */
#if DEBUG
	dbprt("chk","chk - NAMES RECORD\n") ;
#endif
	Check() ;

	Putc(SEGDEF) ;			/* SEGMENT DEFINITION record */
	Putw(7) ;			/* segment record size */
	Putc(0x48) ;			/* word aligned */
	Putw(Text) ;			/* text segment length */
	Putc(TEXTN) ;			/* text segment index */
	Putc(5) ;			/* class name CODE */
	Putc(ovnindx) ;			/* empty overlay name */
#if DEBUG
	dbprt("chk","chk - TEXT SEGMENT RECORD\n");
#endif
	Check() ;			/* checksum and output */

	RECPTR = RECSTART+4;
	RECCNT = 4 ;			/* skip to size field */
	Putw(Data) ;			/* data segment length */
	Putc(DATAN) ;			/* data index */
	Putc(6) ;			/* class name DATA */
	Putc(ovnindx) ;			/* empty overlay name */
#if DEBUG
	dbprt("chk","chk - DATA SEGMENT RECORD\n") ;
#endif
	Check() ;			/* checksum and output */

	RECPTR = RECSTART+4;
	RECCNT = 4 ;			/* position to attributes */
	Putw(Bss) ;			/* length of bss segment */
	Putc(BSSN) ;			/* bss index */
	Putc(7) ;			/* class name BSS */
	Putc(ovnindx) ;			/* empty overlay name */
#if DEBUG
	dbprt("chk","chk - BSS SEGMENT RECORD\n") ;
#endif
	Check() ;			/* checksum and output */

	if(end_out)
		{
		RECPTR = RECSTART+3;
		RECCNT = 3 ;			/* position to attributes */
		Putc(0x48) ;			/* word aligned */
		Putw(0) ;			/* length of end segment */
		Putc(EODN) ;			/* _EOD index */
		Putc(8) ;			/* class name EOD */
		Putc(ovnindx) ;			/* empty overlay name */
#if DEBUG
		dbprt("chk","chk - EOD SEGMENT RECORD\n") ;
#endif
		Check() ;			/* checksum and output */
		}

	Putc(GRPDEF) ;			/* GROUP DEFINITION record */
	if(end_out)
		{			/* 3 segment index,grp index,typ,chks */
		Putw(8) ;
		}
	else
		{
		Putw(6) ;		/* 2 segment index,grp index,typ,chks */
		}
	Putc(GROUPN) ;			/* index to name */
	Putc(0xFF) ;			/* segment list */
	Putc(DATAS);			/* data index */
	Putc(0xFF) ;			/* segment list */
	Putc(BSSS) ;			/* bss index */
	if(end_out) {
		Putc(0xFF) ;		/* segment list */
		Putc(EODS) ;		/* eod index */
	}
#if DEBUG
	dbprt("chk","chk - GROUP RECORD\n") ;
#endif
	Check() ;

#if DEBUG
	dbprt("chk","chk - PUBLIC SYMBOLS RECORD\n") ;
#endif
	dumppsym() ;			/* dump public symbols records */
#if DEBUG
	dbprt("chk","chk - EXTERNAL SYMBOLS RECORD\n") ;
#endif
	dumpesym() ;			/* dump external symbols records */

	Putc(FIXUP) ;			/* define threads for fix up */
	Putw(11) ;			/* 2 * no.of threads + 1 */
	Putc(0) ;			/* target thread 0 */
	Putc(TEXTS) ;			/* text segment */
	Putc(1) ;			/* target thread 1 */
	Putc(DATAS) ;			/* data segment */
	Putc(2) ;			/* target thread 2 */
	Putc(BSSS) ;			/* data segment */

	Putc(0x40) ;			/* frame thread 0 */
	Putc(TEXTS) ;			/* text segment */
	Putc(0x45) ;			/* frame thread 1 */
	Putc(GROUPG) ;			/* dgroup group */
#if DEBUG
	dbprt("chk","chk - THREADS\n") ;
#endif
	Check() ;

	rewind(object);
	Linenum = 1;
	Text = Data = Bss = 0 ;		/* clear location counters */
#if DEBUG
	dbprt("chk","chk - DATA and FIXUP RECORDS\n") ;
#endif
	while ( (typ = getc(object)) != R_END && typ != EOF ) {
#if DEBUG
		dbprt("fix","loc cnt value %x,old value %x,type %02.2x\n",
			   loc,old_loc,typ) ;
#endif
		if ( typ >= R_LINE ) { switch(typ)
			{
			case R_LINE:
				if(Linecnt == LINEMAX)
					dumpline(loctype) ;
					/* ouput the linenumber record */
				Lines[Linecnt++] = ++Linenum;
				Lines[Linecnt++] = loc ;
#if DEBUG
				dbprt("fix","line number changed to %d\n",
				Linenum) ;
#endif
				break;
			case R_TEXT:
				if (loctype != A_TEXT ) {
				/* switch to text location counter */
				old_loc = 
				     switchloc(&locptr,&loctype,&Text,A_TEXT) ;
#if DEBUG
				dbprt("fix","switching to text locat'n cnt\n") ;
#endif
				}
				break;
			case R_DATA:
				if (loctype != A_DATA ) {
				/* switch to data location counter */
				old_loc =
				     switchloc(&locptr,&loctype,&Data,A_DATA) ;
#if DEBUG
				dbprt("fix","switching to data locat'n cnt\n") ;
#endif
				}
				break;
			case R_BSS:
				if (loctype != A_BSS ) {
				old_loc =
				     switchloc(&locptr,&loctype,&Bss,A_BSS) ;
#if DEBUG
				dbprt("fix","switching to bss locat'n cnt\n") ;
#endif
				}
				break;
			case R_FTEXT:
				textfram = 1 ;
				targfram = 0 ;
#if DEBUG
				dbprt("fix","Use text as frame reference\n") ;
#endif
				break;
			case R_FDATA:
				textfram = 0 ;
				targfram = 0 ;
#if DEBUG
				dbprt("fix","Use data as frame reference\n") ;
#endif
				break;
			case R_TARG:
				targfram = 1 ;
				textfram = 0 ;
#if DEBUG
				dbprt("fix","Use target as frame reference\n") ;
#endif
				break;
			case R_SELFR:
				self_rel = 1 ;
#if DEBUG
				dbprt("fix","Self relative fix up\n") ;
#endif
				break;
			case R_SEGR:
				self_rel = 0 ;
#if DEBUG
				dbprt("fix","Seg relative fix up\n") ;
#endif
				break;
			default:
				if(typ == R_LNPTR)
					lngout = 1;
				/* M0001 - change textfram to targfram */
				relsize = typ ;	/* flag long pointers,offsets */
						/* or segment base references */
#if DEBUG
				dbprt("fix","Special relocation type %d\n",
				relsize) ;
#endif
				break;
			}
		} else { 
			switch ( typ & 017 ) {
				case A_ABS:		/* absolute output */
					if ( typ & R_BYT )
						val = getc(object);
					else
						val = getw(object);
					rel = 0;
#if DEBUG
					dbprt("fix","case A_ABS\t") ;
#endif
					break;
				case A_TEXT:		/* resolved first pass*/
				case A_DATA:
				case A_BSS:
					val = getw(object);
					rel = (typ & 017);
#if DEBUG
					dbprt("fix",
					"value %x\trelocation %x\t",val,rel) ;
#endif
					break;
				case A_UND:		/* undefine symbol ref*/
					Sym = (struct Sym *)getw(object);
					evalsym(Sym) ;	/* symbol table ptr */
#if DEBUG
					dbprt("fix",
					"symbol of type %x named %s\n",
					Sym->sy_type,Sym->sy_name) ;
#endif
					if ( Sym->sy_type & A_UEXT )
						val = 0;
					else
						val = Sym->sy_value;
					rel = Sym->sy_type;
					if(!rel && self_rel)
						self_rel = 0;
					break;
				case A_EXPR:		/* expression ref */
					expr = (struct Expr *)getw(object);
					evalexp(expr,&mval,&mrel);
					val = mval;
					rel = mrel;
					if(!rel && self_rel)
						self_rel = 0;
#if DEBUG
					dbprt("fix",
					"expression value %x,type %x\t",
					val,rel) ;
#endif
					break;
			}
			if ( typ & R_PC_REL ) {
#if DEBUG
				dbprt("fix","program counter relative\n") ;
#endif
				if ( rel != loctype && (rel & A_UEXT == 0)) {
					error("Illegal jump out of segment","");
					if ( typ & R_BYT ) {
						loc++;
						fillbuf(1,0,loctype,loc) ;
#if DEBUG
						dbprt("fix",
						"error byte value 0\n") ;
#endif
						} 
					else {
						loc += 2;
						fillbuf(2,0,loctype,loc) ;
#if DEBUG
						dbprt("fix",
						"error word value 0\n") ;
#endif
						}
					rel = 0;
					textfram = 0 ;
					self_rel = 0 ;
					targfram = 0 ;
					}
				else if ( typ & R_BYT ) {
					loc++;
					val -= loc;
					if ( val < -128 || val > 127 )
						error("Short jump out of range",
						"");
					rel = 0;
					fillbuf(1,val,loctype,loc) ;
					textfram = 0 ;
					targfram = 0 ;
					self_rel = 0 ;

#if DEBUG
					dbprt("fix",
					"relative byte value %x\n",val) ;
#endif
				} else {
					loc += 2;
					if(rel & A_UEXT)
						val = 0;
					else
						{
						val -= loc;
						rel = 0;
						textfram = 0 ;
						self_rel = 0 ;
						targfram = 0 ;
						}
					fillbuf(2,val,loctype,loc) ;
#if DEBUG
					dbprt("fix",
					"relative word value %x\n",val) ;
#endif
				}
			} else {
				if ( typ & R_BYT ) {
					if(relsize)
						{
						error("byte segment reference",
						"")  ;
						relsize = rel = 0 ;
						}
					loc++;
					fillbuf(1,val,loctype,loc) ;
#if DEBUG
					dbprt("fix",
					"output byte value %x\n",val) ;
#endif
				} else {
					loc += 2;
					fillbuf(2,val,loctype,loc) ;
#if DEBUG
					dbprt("fix",
					"output word value %x\n",val) ;
#endif
				}
			}
			if ( loc-old_loc > 0 ) {	/* something output */
#if DEBUG
				dbprt("fix","relocation value %x\n",rel) ;
#endif
				if(rel && rel != 3)	/* not abs or undef */
					{
					if(rel & A_UEXT)
						{	/* external reference */
						FIX.target = F_EXT;
#if DEBUG
						dbprt("fix","ext ref no# %d\n",
						Sym->sy_snum) ;
#endif
						FIX.index = Sym->sy_snum;
						if(!targfram)
						    {
#if DEBUG
						    dbprt("fix","!textfram\n");
#endif
						    FIX.frame=textfram ? 
						    FTXTTHRD:FDGPTHRD;
						    FIX.frame |= 0x80;
						    }
						else	
						    {
#if DEBUG
						    dbprt("fix","targfram\n") ;
#endif
						    FIX.frame=0x50;
						    }
						}
					else
					    {
					    if(!targfram)

					    	{
						if(!textfram)
						    {
						    FIX.target = (rel==A_TEXT) ?
						    TTXTTHRD : ((rel==A_DATA) ?
						    TDATTHRD : TBSSTHRD) ;
						    FIX.frame = (rel==A_TEXT) ?
						    FTXTTHRD : FDGPTHRD ;
						    }
						else
						    {
						    FIX.target = TTXTTHRD;
						    FIX.frame  = FTXTTHRD;
						    }
						FIX.frame |= 0x80;
						}
					    else
					        {
					        FIX.frame=0x50;
					        if(!textfram)
						    FIX.target = (rel==A_TEXT) ?
						    TTXTTHRD : ((rel==A_DATA) ?
						    TDATTHRD : TBSSTHRD) ;
					        else
						    FIX.target = TTXTTHRD;
					        }
					    }
#ifdef DEBUG
					dbprt("fix","self_rel %d\n",self_rel);
#endif
					if(self_rel)
						reltyp = 0x8000;
					else
						reltyp = 0xc000;
#ifdef DEBUG
					dbprt("fix","reltyp %x\n",reltyp);
#endif
					FIX.loc_addr = reltyp | 
						(old_loc-lastloc[loctype]);
					if(relsize > R_FTEXT)
						{	/* segment reference */
						FIX.loc_pos = relsize - R_FTEXT;
						relsize = 0 ;
						}
					else
						FIX.loc_pos = ((rel & R_BYT) ?
						  0 :	/* low byte */
						  1);	/* offset */
					if (++FIXCNT >= FIXREC)
						{
#ifdef DEBUG
						dbprt("rel",
						"call dumpbuf,FIXCNT>FIXREC\n");
#endif
						dumpbuf(loctype,loc) ;
						}
					self_rel = textfram = targfram = 0 ;
					}
				old_loc = loc ; 	/* remember last loc */
				relsize = 0;		/* kill special rel */
			}
		}
	}
	if(RECCNT > DATAHD)
		dumpbuf(loctype,loc) ;			/* clear buffers */
	dumpline(loctype) ;
	RECCNT = 0 ;				/* reinitialize buffers */
	RECPTR = RECSTART ;
	Putc(MODEND) ;
	Sym = lookup("start") ;
	if(Sym->sy_type & A_TEXT && (Sym->sy_flags & F_EXTRN))
		{
		length = 5;
		Putw(length) ;
		Putc(0xc1) ;			/* main module, logical addr */
		Putc(0x8c) ;			/* frame text,target text */
		Putw(Sym->sy_value) ;		/* starting address */
		}
	else
		{
		length = 2 ;
		Putw(length) ;
		Putc(1) ;			/* logical address */
		}
	Check() ;
}

#if DEBUG
obj_list()
	{
	register int typ,Linenum ;
	rewind(object);
	Linenum = 1;

	while ( (typ = getc(object)) != R_END && typ != EOF ) 
		{
		if(typ == R_LINE)
			{		/* could match erroneously */
			Linenum++;
			dbprt("obj","\n\nLine number %d\n",Linenum) ;
			}
		dbprt("obj","%x ",typ) ;
		}
	}
#endif

evalsym(sym)
register struct Sym *sym;
{
	register int tmp;
	int val,type;

	if ( sym->sy_type == A_EXPR ) {
					/* symbol value is in expression form */
		if ( sym->sy_flags & F_BUSY ) {
			error("Reference loop in definition of symbol ",
			       sym->sy_name);
			return (-1);
		} else {		/* not self referencing */
			sym->sy_flags |= F_BUSY;
			tmp = evalexp((struct Expr *)sym->sy_value,&val,&type);
					/* evaluate symbol expression */
			sym->sy_flags &= -F_BUSY - 1;
			if ( tmp == -1 ) 
				{	/* error in evaluation */
				sym->sy_value = 0;
				sym->sy_type = A_ABS;
				return (-1);
			} else {	/* save value and type */
				sym->sy_value = val;
				sym->sy_type = type;
			}
		}
	} else if ( sym->sy_type == A_UND ) {
		if ( (sym->sy_flags & F_LOCAL) == 0 ) {	
					/* not local */
			if( Und_glb ) {	/* treat undefines as gloabl */
				if ( (sym->sy_flags & F_EXTRN) == 0 )	
					/* not external - define as global */
					do_glb(sym);
				sym->sy_type = A_UEXT ;
			} else
				error("Undefined symbol ",sym->sy_name);
		} else {				/* undefined local */
			fprintf(list,"Undefined local symbol #%d, line %d!\n",
			sym->sy_nam.sy_int,Linenum);
			++Errors;
		}
	}
	return (0);
}

evalexp(expr,val,type)
register struct Expr *expr; register int *val,*type;
{
	if ( getval(&(expr->E_a)) == -1 || getval(&(expr->E_b)) == -1 ) {
		if ( Pass == 1 )		/* possible forward reference */
			return (1);		/* ok for first pass */
		else {
			*val = 0;		/* not evaluatable  use zero */
			*type = A_ABS;
			return (-1);		/* return error flag */
		}
	}
	switch ( expr->E_opr ) {
		case '+':
			*val = expr->E_a.e_v.e_int + expr->E_b.e_v.e_int;
			if ( expr->E_a.e_type == A_ABS )
				*type = expr->E_b.e_type;
			else
				if ( expr->E_b.e_type == A_ABS )
					*type = expr->E_a.e_type;
				else {
					error("Illegal relocatable expression","");
					return (-1);
				}
			if((*type==A_TEXT) || (*type==A_DATA) || (*type==A_BSS))
				Fix_it = 1;	/* fixed core location */
			break;
		case '-':
			*val = expr->E_a.e_v.e_int - expr->E_b.e_v.e_int;
			if ( expr->E_a.e_type == A_ABS || (expr->E_b.e_type & A_UEXT) )
				if ( expr->E_b.e_type == A_ABS )
					*type = expr->E_a.e_type;
				else {
					error("Illegal relocatable expression","");
					return (-1);
				}
			else
				if ( expr->E_b.e_type == A_ABS )
					*type = expr->E_a.e_type;
				else if ( expr->E_b.e_type == expr->E_a.e_type )
					*type = A_ABS;
				else {
					error("Illegal relocatable expression","");
					return (-1);
				}
			break;
		case '*':
			*val = expr->E_a.e_v.e_int * expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '/':
			*val = expr->E_a.e_v.e_int / expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '&':
			*val = expr->E_a.e_v.e_int & expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '^':
			*val = expr->E_a.e_v.e_int % expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '!':
			*val = expr->E_a.e_v.e_int | expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case RSH:
			*val = expr->E_a.e_v.e_int >> expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case LSH:
			*val = expr->E_a.e_v.e_int << expr->E_b.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS || expr->E_b.e_type != A_ABS) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
		case '~':
			*val = ~ expr->E_a.e_v.e_int;
			*type = A_ABS;
			if ( expr->E_a.e_type != A_ABS ) {
				error("Illegal relocatable expression","");
				return (-1);
			}
			break;
	}
	return (0);
}

getval(estk)
register struct estak *estk;
{
	register struct Sym *sym;
	int self,val,type,base,offset,typ;

	base = estk->e_type & E_BASE ;		/* base of addr requested */
	offset = estk->e_type & E_OFFS ;	/* offset of addr requested */
	self = estk->e_type & E_SELF ;		/* self rel addr requested */
	typ = estk->e_type &= 0xff ;		/* strip off base and offset */
	if ( Pass == 1 ) {
		if ( typ == A_UND || typ == A_EXPR )
			return (-1);	/* can't evaluate yet */
	} else if ( typ == A_UND ) {
		sym = estk->e_v.e_sym;	/* still undefined ? */
		if(evalsym(sym) == -1 )	/* try evaluating symbol again */
			return (-1);	/* not possible to evaluate */
		else {
			estk->e_type = sym->sy_type | base | offset | self;
			if ( sym->sy_type & A_UEXT )
				{	/* if external use zero */
				estk->e_v.e_int = 0;
				Sym = sym ;
				}
			else
					/* otherwise use symbol value */
				estk->e_v.e_int = sym->sy_value;
		}
	} else if ( typ == A_EXPR ) {
		if ( evalexp(estk->e_v.e_expr,&val,&type) == -1 )
			return (-1);	/* un evaluatable expression */
		else {
			estk->e_type = type;
			estk->e_v.e_int = val;
		}
	}
	return (0);
}

dumppsym()			/* dump public symbols */
	{
	register struct Sym *sym;
	register int length ;

	for ( sym = gl_head ; sym ; sym = sym->sy_glink ) 
		{
#if DEBUG
			dbprt("sym",
			"pubsym -> ? %s,value %x,type %x,flags %x\n",
			sym->sy_name,sym->sy_value,sym->sy_type,sym->sy_flags) ;
#endif
		evalsym(sym);                   /* evaluate the symbol */
		if ((sym->sy_flags & F_EXTRN) == 0)
			continue;               /* only externals eligible */
		if ((sym->sy_type != A_TEXT) && (sym->sy_type != A_DATA) &&
		    (sym->sy_type != A_BSS)  && (sym->sy_type != A_ABS))
			continue;               /* not our type */

		if (((sym->sy_flags&F_FIXED)==0) && (sym->sy_type != A_ABS)){
			printf("%s Ia86 internal error: bad SYM format\n",
			sym->sy_name);
			exit(1);
		}
#if DEBUG
		dbprt("sym",
		"sym -> global %s,value %x,type %x,flags %x\n",
		sym->sy_name,sym->sy_value,sym->sy_type,sym->sy_flags) ;
#endif
		Putc(0x90) ;            /* Pubilc Name Record */
		if(strlen(sym->sy_name) > 8)
			*(sym->sy_name+8) = 0 ;
		length = strlen(sym->sy_name) +
			((sym->sy_type == A_ABS) ? 9:7);
		Putw(length) ;
		switch (sym->sy_type) {
		case A_DATA:
			Putc(GROUPG);           /* group index */
			Putc(DATAS);            /* data segment */
			break;
		case A_TEXT:
			Putc(0);                /* no group */
			Putc(TEXTS);            /* text segment */
			break;
		case A_BSS:
			Putc(GROUPG);           /* group index */
			Putc(BSSS);             /* BSS segment */
			break;
		case A_ABS:
			Putc(0);                /* no group */
			Putc(0);                /* no segment */
			Putw(0);                /* off of frame 0 */
		}
		Putc(strlen(sym->sy_name)) ;
		Puts(sym->sy_name) ;     /* output the name */
		Putw(sym->sy_value) ;    /* output its offset */
		Putc(0) ;               /* no type index */
		Check() ;               /* evaluate checksum */
		}
	if(end_out)
		{
		Putc(0x90) ;		/* Pubilc Name Record */
		length = 4 + 7 ;	/* _end + info */
		Putw(length) ;
		Putc(GROUPG) ;		/* data group */
		Putc(EODS) ;		/* end segment */
		Putc(4) ;
		Puts("_end") ;		/* output the name */
		Putw(0) ;		/* output its offset */
		Putc(0) ;		/* no type index */
		Check() ;		/* evaluate checksum */
		}
	}
static char typindx[256] ;
dumpesym()			/* dump external symbols */
	{
	register struct Sym *sym;
	register int i ;
	int length ;
	long numbits ;
	int typdef,typcnt = 0 ;
	int count,temp ;
	char ctemp;

	i = 1 ;
	for ( sym = gl_head ; sym ; sym = sym->sy_glink ) 
		{
#if DEBUG
			dbprt("sym",
			"sym -> ? %s,value %x,type %x,flags %x\n",
			sym->sy_name,sym->sy_value,sym->sy_type,sym->sy_flags) ;
#endif
		if((sym->sy_type & A_UEXT) &&
		  (sym->sy_flags & F_EXTRN))
			{
#if DEBUG
			dbprt("sym",
			"sym -> external %s,value %x,type %x,flags %x\n",
			sym->sy_name,sym->sy_value,sym->sy_type,sym->sy_flags) ;
#endif
			if(sym->sy_value)
			    {
			    numbits = (long)sym->sy_value * 8 ;
						/* convert to bits */
			    if(numbits > MAXCOMMON)
			    	{
			    	error("common variable %s too large",
			    	sym->sy_name) ;
			    	numbits = MAXCOMMON;
			    	}
			    if ((sym->sy_value < 256) && typindx[sym->sy_value])
				{
			    	typdef = typindx[sym->sy_value] ;
#ifdef DEBUG
				dbprt("typ","using old typdef of %d\n",typdef);
#endif
				}
			    else
				{
				Putc(TYPDEF) ;
				typdef = ++typcnt ;
				if(numbits < 128)
					count = 6 ;
				else if (numbits < 65536L)
					count = 8 ;
				else
					count = 9 ;
				Putw(count) ;
				temp = 0 ;
				Putw(temp) ;
				Putc(0x62) ;
				Putc(0x7b) ;
				if(count == 8)
					{
					Putc(0x81) ;
					}
				else if(count == 9)
					{
					Putc(0x84) ;
					} ;
				if (count == 6)
					{
					ctemp = numbits ;
					Putc(ctemp) ;
					}
				else if (count == 8)
					{
					temp = numbits;
					Putw(temp) ;
					}
				else if (count == 9)
					{
					temp = numbits;
					Putw(temp) ;
					ctemp = numbits >> 16 ;
					Putc(ctemp) ;
					}
#ifdef DEBUG
	dbprt("chk","chk - TYPDEF COMMON RECORD\n") ;
#endif
				Check();	/* output and checksum */
				if(sym->sy_value < 256)
					typindx[sym->sy_value] = typdef;
				}
			    }
			else
				typdef = 0 ;
			Putc(EXTNAMS) ;
			if(strlen(sym->sy_name) > 8)
				*(sym->sy_name+8) = 0 ;
			length = strlen(sym->sy_name) + 3 ;
			if(typdef > 127)
				length++ ;	/* large typdef index */
			Putw(length) ;
			Putc(strlen(sym->sy_name)) ;
			Puts(sym->sy_name) ;
			if(typdef <= 127)
				{		/* small index */
				Putc(typdef) ;
				}
			else
				{		/* large index */
				typdef = 0x8000 | typdef ;
				Putw(typdef) ;
				}
#ifdef DEBUG
	dbprt("chk","chk - EXTERN OR COMMON SYMBOL RECORD\n") ;
#endif
			Check() ;
			sym->sy_snum = i++ ;	/* index into global */
			}
		}
	}
