    /****************************************************************
    *                                                               *
    *                           NEWSYM.C                            *
    *                                                               *
    *  Symbol table routines.                                       *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types and constants */
#include                "bndtrn.h"      /* More types and constants */
#include                "bndvrt.h"      /* Virt. Mem. extern declarations */
#include                "lnkio.h"       /* Linker I/O definitions */

extern                  Fatal();        /* Fatal error procedure */
extern BYTE             *ObGetVa();     /* Fetch from virtual memory */

extern FTYPE            rgfDirty[];     /* Used by DirtyBp() macro */
extern WORD             vbp;            /* Virtual block pointer */
extern char             *vob;           /* Virtual object */
extern RBTYPE           rgrhte[];       /* Array of Real Hash Table Entries */
extern WORD             rbMacSyms;      /* Ptr. to first free word in area */
extern FTYPE            vfCreated;      /* "VirtobjFlagCREATED" */
extern FTYPE            fIgnoreCase;    /* Ignore case of alpha char.s */
extern HTETYPE          vhte;           /* Virtual hash table entry */
extern RBTYPE           vrprop;
extern RBTYPE           vrhte;
static WORD             mpattrcb[] =    /* Map attribute to struct size */
			  {
			    3,
			    CBPROPSN,   /* Size of APROPSNTYPE */
			    CBPROPSN,   /* Size of APROPSNTYPE */
			    CBPROPNAME, /* Size of APROPNAMETYPE */
			    CBPROPNAME, /* Size of APROPNAMETYPE */
			    CBPROPFILE, /* Size of APROPFILETYPE */
			    CBPROPGROUP,/* Size of APROPGROUPTYPE */
			    CBPROPRAGSN,/* Size of APROPRAGSNTYPE */
			    CBPROPUNDEF,/* Size of APROPUNDEFTYPE */
			    CBPROPOVERLAY,
					/* Size of APROPOVERLAYTYPE */
			    3
			  };

    /****************************************************************
    *                                                               *
    *  IHashKey:                                                    *
    *                                                               *
    *  This function hashes a  length-prefixed  string and returns  *
    *  hash value.                                                  *
    *                                                               *
    ****************************************************************/

LOCAL WORD              IHashKey(psb)
BYTE                    *psb;           /* Pointer to length-prefixed string */
  {
    register WORD       i;              /* Index */
    register WORD       hashval;        /* Hash value */

    /* NOTE:  IRHTEMAX MUST BE 256 OR THIS FUNCTION WILL FAIL */
    i = B2W(psb[0]);                    /* Get length */
    hashval = i;                        /* Use length as initial hash value */
    while(i)                            /* While not at beginning of string */
      {
	hashval = ror(hashval,2,BYTELN) ^ (B2W(psb[i--]) | 040);
					/* Hash */
      }
    return(hashval);                    /* Return value */
  }

    /****************************************************************
    *                                                               *
    *  RbAllocSymNode:                                              *
    *                                                               *
    *  This function takes as its input a WORD n, and it allocates  *
    *  n  continguous  bytes  in  the  symbol  area  and returns a  *
    *  pointer  to  the  first  of  those  bytes.  The  bytes  are  *
    *  guaranteed to reside on the same  virtual page (when not in  *
    *  memory).                                                     *
    *                                                               *
    ****************************************************************/

LOCAL RBTYPE            RbAllocSymNode(cb)
WORD                    cb;             /* Number of bytes to allocate */
  {
    RBTYPE              rbLimVbf;       /* End of current page */
    RBTYPE              rb;             /* Address of allocated bytes */

    ASSERT(cb <= CBVBF);                /* Cannot alloc. more than 512 bytes */
    cb = (cb + 1) & ~1;                 /* Word align (PDP-11 constraint) */
    rbLimVbf = (rbMacSyms + CBVBF - 1) & ~(CBVBF - 1);
					/* Find limit of current page */
    if(rbMacSyms + cb > rbLimVbf) rbMacSyms = rbLimVbf;
					/* If alloc. would cross page
					*  boundary, start with new page
					*/
    rb = rbMacSyms;                     /* Get address to return */
    rbMacSyms += cb;                    /* Update pointer to end of area */
    if(rbMacSyms >> 9 >= CPNMAXSYMS)
      Fatal("Symbol table capacity exceeded.",0x80);
					/* Check if page limit exceeded */
    vob = ObGetVa(AREASYMS,rb,TRUE);    /* Allocate space in virt mem */
    return(rb);                         /* Return the address */
  }

    /****************************************************************
    *                                                               *
    *  PropAdd:                                                     *
    *                                                               *
    *  This function  adds a property to a  hash table entry node.  *
    *  It returns the  location of  the property.  Inputs  are the  *
    *  virtual  address of the hash table  entry and the attribute  *
    *  (or property) to be added.                                   *
    *                                                               *
    ****************************************************************/

PROPTYPE                PropAdd(rhte,attr)
RBTYPE                  rhte;           /* Virtual addr of hash tab ent */
ATTRTYPE                attr;           /* Attribute to add to entry */
  {
    AHTETYPE            *hte;           /* Hash table entry pointer */
    APROPTYPE           *aprop;         /* Property list pointer */
    BPTYPE              bp;             /* Block number */
    WORD                i;              /* Index variable */

    hte = (AHTETYPE *) ObGetVa(AREASYMS,rhte,FALSE);
					/* Fetch from VM */
    bp = vbp;                           /* Save block number */
    vrprop = RbAllocSymNode(mpattrcb[attr]);
					/* Allocate symbol space */
    for(i = 0; i < mpattrcb[attr]; ++i) vob[i] = '\0';
					/* Zero the space */
    aprop = (APROPTYPE *) vob;          /* Recast pointer */
    aprop->a_attr = attr;               /* Store the attribute */
    aprop->a_next = hte->rprop;         /* Set link */
    hte->rprop = vrprop;                /* Complete link */
    DirtyBp(bp);                        /* Mark page as changed */
    return((PROPTYPE) aprop);           /* Return pointer */
  }

    /****************************************************************
    *                                                               *
    *  PropRhteLookup:                                              *
    *                                                               *
    *  "Look up a property on hash  table entry (possibly creating  *
    *  it) and return pointer to property.                          *
    *                                                               *
    *  Input:   rhte    Virtual address of hash table entry.        *
    *           attr    Property to look up.                        *
    *           fCreate Flag  to  create   property  cell  if  not  *
    *                   present.                                    *
    *  Return:          pointer to property cell."                  *
    *                                                               *
    ****************************************************************/

PROPTYPE                PropRhteLookup(rhte,attr,fCreate)
RBTYPE                  rhte;           /* Virt. addr. of hash table entry */
ATTRTYPE                attr;           /* Property to look up */
FTYPE                   fCreate;        /* Create property cell flag */
  {
    APROPTYPE           *aprop;
    AHTETYPE            *ahte;

    vrhte = rhte;                       /* Set global */
    vhte = ObGetVa(AREASYMS,rhte,FALSE);
					/* Fetch from VM */
    ahte = (AHTETYPE *) vhte;           /* Recast pointer */
    vrprop = ahte->rprop;
    vfCreated = FALSE;                  /* Assume no creation takes place */
    for(;;)
      {
	aprop = (APROPTYPE *) ObGetVa(AREASYMS,vrprop,FALSE);
					/* Fetch from VM */
	if(aprop->a_attr == attr) return((PROPTYPE) aprop);
					/* Return if match found */
	vrprop = aprop->a_next;         /* Try next item in list */
	if(aprop->a_attr == ATTRNIL)    /* If no entry here */
	  {
	    if(!fCreate) return(PROPNIL);
					/* Return nil if creation not wanted */
	    vfCreated = TRUE;           /* A new creation! */
	    return(PropAdd(vrhte,attr));/* Let someone else do it */
	  }
      }
  }

    /****************************************************************
    *                                                               *
    *  PropSymLookup:                                               *
    *                                                               *
    *  This  function  looks  up  a  symbol  and its property.  It  *
    *  can create an entry, if necessary.  It returns a pointer to  *
    *  the property cell.  It takes  as its inputs  a pointer to a  *
    *  symbol, the property to look up, and a flag which specifies  *
    *  if a new  property cell is to be  created in the event that  *
    *  one of the given type does not already exist.                *
    *                                                               *
    ****************************************************************/

PROPTYPE                PropSymLookup(psym,attr,fCreate)
BYTE                    *psym;          /* Pointer to length-prefixed string */
ATTRTYPE                attr;           /* Attribute to look up */
FTYPE                   fCreate;        /* Create prop cell if not found */
  {
    WORD                irhte;          /* Hash table index */
    AHTETYPE            *ahte;          /* Pointer to hash table entry */

    ENTER(PropSymLookup);               /* Debug info */
#if DEBUG                               /* If debugging on */
    fprintf(stderr,"  Looking up ");    /* Message */
    OutSb(stderr,psym);                 /* Symbol */
    fprintf(stderr,"(%d) with attr %d\n",B2W(psym[0]),B2W(attr));
#endif                                  /* End debugging code */
    irhte = IHashKey(psym);             /* Get hash index */
    vrhte = rgrhte[irhte];
    vfCreated = FALSE;                  /* Assume nothing will be created */
    for(;;)
      {
	if(vrhte == RHTENIL)            /* If nil pointer */
	  {
	    DEBUGMSG("  Empty slot found");
					/* Debug message */
	    if(!fCreate) return(PROPNIL);
					/* Return nil if creation disallowed */
	    vfCreated = TRUE;           /* New creation */
	    vrhte = RbAllocSymNode(CBHTE + B2W(psym[0]));
					/* Allocate space for symbol entry */
	    vhte = vob;                 /* Save virtual address */
	    ahte = (AHTETYPE *) vhte;   /* Recast pointer */
	    ahte->rhteNext = rgrhte[irhte];
	    ahte->attr = ATTRNIL;       /* No attributes yet */
	    ahte->rprop = vrhte;        /* Add to list */
	    Move(B2W(psym[0]) + 1,psym,ahte->cch);
					/* Load symbol */
	    rgrhte[irhte] = vrhte;
	    LEAVE(PropSymLookup,-1L);   /* Debug info */
	    if(attr != ATTRNIL) return(PropAdd(vrhte,attr));
					/* Add the attribute */
	    return(PROPNIL);            /* Nothing to return */
	  }
	DEBUGMSG("  Collision");        /* Debug message */
	vhte = ObGetVa(AREASYMS,vrhte,FALSE);
					/* Fetch from VM */
	ahte = (AHTETYPE *) vhte;       /* Recast pointer */
	if(!SbCompare(psym,ahte->cch,fIgnoreCase))
	  return(PropRhteLookup(vrhte,attr,fCreate));
					/* If symbol found, get property */
	vrhte = ahte->rhteNext;         /* Move down list */
      }
  }

    /****************************************************************
    *                                                               *
    *  The legendary EnSyms:                                        *
    *                                                               *
    *  This  function  applies  a function to  all symbol/property  *
    *  pairs that  have a  particular  property.  It takes  as its  *
    *  inputs  a pointer to a function  to call and a  property to  *
    *  look for.  EnSyms() does not return a meaningful value.      *
    *                                                               *
    ****************************************************************/

EnSyms(pproc,attr)
int                     (*pproc)();     /* Function pointer */
ATTRTYPE                attr;           /* Attribute to key on */
  {
    WORD                irhte;          /* Hash table index */
    RBTYPE              rhte;           /* Hash table entry address */
    ATTRTYPE            attrT;
    APROPTYPE           *aprop;         /* Pointer to a property cell */
    AHTETYPE            *ahte;          /* Pointer to a hash table entry */
    FTYPE               fNewHte;
    RBTYPE              rprop;
    RBTYPE              rhteNext;
    RBTYPE              rpropNext;

    ENTER(EnSyms);                      /* Debug info */
    for(irhte = 0; irhte < IRHTEMAX; ++irhte)
      {                                 /* Look through hash table */
	rhte = rgrhte[irhte];           /* Get pointer to chain */
	while(rhte != RHTENIL)          /* While not at end of chain */
	  {
	    ahte = (AHTETYPE *) ObGetVa(AREASYMS,rhte,FALSE);
					/* Fetch entry from VM */
	    fNewHte = TRUE;             /* First call on this hash tab entry */
	    rhteNext = ahte->rhteNext;  /* Get pointer to next in chain */
	    rprop = ahte->rprop;        /* Get pointer to property list */
	    for(;;)                     /* Loop to search property list */
	      {
		aprop = (APROPTYPE *) ObGetVa(AREASYMS,rprop,FALSE);
					/* Fetch entry from VM */
		rpropNext = aprop->a_next;
					/* Get pointer to next in list */
		attrT = aprop->a_attr;  /* Get the attribute */
		if(attr == attrT || attr == ATTRNIL)
		  (*pproc)(ahte,aprop,rhte,rprop,fNewHte);
					/* If prop acceptable, apply func */
		if(attrT == ATTRNIL) break;
					/* Break if at end of prop list */
		fNewHte = FALSE;        /* Next call (if any) won't be first */
		rprop = rpropNext;      /* Move down the list */
	      }
	    rhte = rhteNext;            /* Move down the chain */
	  }
      }
    LEAVE(EnSyms,0xFFFF);               /* Debug info */
  }
