    /****************************************************************
    *                                                               *
    *                           NEWTP2.C                            *
    *                                                               *
    *  This module contains functions to process object modules on  *
    *  pass 2.                                                      *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types, constants, macros */
#include                "bndtrn.h"      /* Types and constants */
#include                "bndvrt.h"      /* Virtual memory module def'ns */
#include                "xenfmt.h"      /* Executable file header def.s */
#include                "lnkio.h"       /* Linker I/O definitions */

#if NOT OIAPX286
extern                  RecordSegmentReference();
					/* Record reference to a segment */
#endif
extern BYTE             Gets();         /* Read a byte of input */
extern WORD             WGets();        /* Read a WORD of input */
extern WORD             GetIndex();     /* Read var. length index from input */
extern                  GetBytes();     /* Read n bytes from input */
extern                  SkipBytes();    /* Skip n bytes of input */
extern                  ErrHlt();       /* Internal error routine */
extern                  Fatal();        /* Fatal error routine */
extern                  LNmRc1();       /* Processes LNAMES records */
extern                  DataRec();      /* Processes data records */
extern                  ComRc1();       /* Processes COMENT records */
extern                  EndRec();       /* Processes MODEND records */
extern                  RelocRec();     /* Processes FIXUPP records */
extern                  ModRc1();      /* Processes THEADRs and LHEADRs */
extern PROPTYPE         PropSymLookup();/* Look up property/symbol pair */
extern PROPTYPE         PropRhteLookup();
					/* Look up property/address pair */
extern PROPTYPE         PropAdd();      /* Add property to attribute list */
extern                  InvalidObject();/* Error message routine */

extern FTYPE            rgfDirty[];     /* Used by DirtyBp() macro */
extern WORD             cbRec;          /* No. of bytes left in obj rec */
extern RECTTYPE         rect;           /* Type of current obj rec */
extern WORD             vbp;            /* Current virtual page number */
extern BSTYPE           bsLst;          /* File handle for map file */
extern BSTYPE           bsInput;        /* File handle for input file */
extern MITYPE           rgmi[];         /* Input file record buffer */
extern WORD             snMac;          /* Count of local SEGDEFs */
extern WORD             extMac;         /* Count of PUBDEFs */
extern WORD             cErrors;        /* Count of non-fatal errors */
extern WORD             gsnMac;         /* Count of global SEGDEFs */
extern WORD             lnameMac;       /* Count of LNAMEs */
extern RBTYPE           vrpropFile;
extern RBTYPE           vrhte;          /* Address of hash table entry */
extern FTYPE            vfLineNos;      /* Line-numbers-requested flag */
extern SNTYPE           vgsnLineNosPrev;
extern WORD             vcln;
extern SEGTYPE          mpgsnseg[];
extern RATYPE           mpgsndra[];
extern RBTYPE           mpgsnrprop[];   /* f(glob SEGDEF) = prop list addr */
extern RATYPE           mpextra[];      /* f(PUBDEF no.) = public offset */
extern SNTYPE           mpextgsn[];     /* f(PUBDEF no.) = glob SEGDEF no. */
extern SNTYPE           mpsngsn[];      /* f(SEGDEF no.) = glob SEGDEF no. */
extern RBTYPE           mplnamerhte[];  /* f(LNAME no.) = hash table address */
extern SATYPE           mpsegsa[];      /* f(seg) = sa */
extern FTYPE            mpsegFlags[];   /* f(seg) = Flags */
#if OIAPX286
extern RATYPE           mpsegraFirst[]; /* f(seg) = offset of 1st byte */
#endif
extern FTYPE            fUndefinedExterns;
					/* Undefined externals flag */
extern GRTYPE           mpgrggr[];      /* f(GRPDEF no.) = global GRPDEF no. */
extern WORD             grMac;          /* Count of GRPDEFs */
extern RECTTYPE         vrectData;      /* Data record type */
extern SEGTYPE          vsegCur;        /* Current segment number */
extern RATYPE           vraCur;         /* Current segment offset */
extern WORD             vcbData;        /* Size of data record */

    /****************************************************************
    *                                                               *
    *  SegRec2:                                                     *
    *                                                               *
    *  This function processes SEGDEF records on pass 2.            *
    *  See pp. 32-35 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

SegRec2()
  {
    TYSNTYPE            tysn;           /* ACBP field */
    ALIGNTYPE           align;          /* Alignment subfield of ACBP */
    SATYPE              saAbs;          /* Frame number of absolute LSEG */
    WORD                seglen;         /* Segment length */
    LNAMETYPE           lname;          /* Segment name index */
    LNAMETYPE           lnameClass;     /* Segment class name index */
    WORD                comb;           /* Segment combining information */
    APROPSNTYPE         *apropSn;       /* Pointer to segment hash tab entry */
    SNTYPE              gsn;            /* Global SEGDEF no. */
    RATYPE              dra;

    ENTER(SegRec2);                     /* Debug info */
    if(gsnMac >= GSNMAX || snMac >= SNMAX) Fatal("Too many segments.",0xA0);
					/* Check for table overflow */
    tysn = Gets();                      /* Read in ACBP byte */
    align = (tysn >> 5) & 7;            /* Get alignment subfield */
    ASSERT(align != 5);                 /* Unnamed absolute not supported */
    ASSERT(align != 6);                 /* LTL LSEG not supported */
    if(align == ALGNABS)                /* If absolute LSEG */
      {
	saAbs = WGets();                /* Read in frame number */
	Gets();                         /* Skip frame offset */
      }
    seglen = WGets();                   /* Read in segment length */
    lname = GetIndex(1,lnameMac - 1);   /* Get segment name index */
    lnameClass = GetIndex(1,lnameMac - 1);
					/* Get segment class name index */
    GetIndex(0,lnameMac - 1);           /* Skip overlay name index */
    switch(align)                       /* Switch on alignment type */
      {
	case ALGNABS:                   /* Absolute LSEG */
	case ALGNWRD:                   /* Word-aligned LSEG */
	case ALGNBYT:                   /* Byte-aligned LSEG */
	case ALGNPAR:                   /* Paragraph-aligned LSEG */
	case ALGNPAG:                   /* Page-aligned LSEG */
	  break;

	default:                        /* Unsupported or illegal types */
	  ASSERT(0 != 0);               /* Should never happen! */
	  mpsngsn[snMac++] = 0;         /* Map to nothing */
	  LEAVE(SegRec2,0xFFFE);        /* Debug info */
	  return;                       /* And return */
      }
    comb = (tysn >> 2) & 7;             /* Get combining info */
    if(comb)                            /* If not a private segment */
      {
	apropSn = (APROPSNTYPE *)
	  PropRhteLookup(mplnamerhte[lname],ATTRPSN,FALSE);
					/* Look up property cell */
	while(apropSn != PROPNIL &&
	  apropSn->as_rCla != mplnamerhte[lnameClass])
	  {                             /* Look for matching class */
	    apropSn = (APROPSNTYPE *)
	      ObGetVa(AREASYMS,apropSn->as_next,FALSE);
					/* Try next link in list */
	  }
      }
    else                                /* Else if private segment */
      {
	apropSn = (APROPSNTYPE *)
	  PropRhteLookup(mplnamerhte[lname],ATTRLSN,FALSE);
					/* Look up property cell */
	while(apropSn != PROPNIL &&
	  apropSn->as_rFil != vrpropFile)
	  {                             /* Look for matching virt address */
	    apropSn = (APROPSNTYPE *)
	      ObGetVa(AREASYMS,apropSn->as_next,FALSE);
					/* Try next link in list */
	  }
      }
    if(apropSn == PROPNIL) ErrHlt(0xA1);/* Should never happen */
    gsn = apropSn->as_gsn;              /* Get global SEGDEF no. */
    if(comb == COMBSTK) mpgsndra[gsn] = apropSn->as_cbMx - seglen;
    else if(comb != COMBCOM)            /* Else if not common comb. */
      {
	dra = mpgsndra[gsn] + apropSn->as_cbPv;
	switch(align)                   /* Switch on alignment type */
	  {
	    case ALGNWRD:               /* Word-aligned LSEG */
	      mpgsndra[gsn] = 0xFFFE & (dra + 1);
					/* Round to next word offset */
	      break;

	    case ALGNPAR:               /* Paragraph-aligned LSEG */
	      mpgsndra[gsn] = 0xFFF0 & (dra + 0xF);
					/* Round to next paragraph offset */
	      break;

	    case ALGNPAG:               /* Page-aligned LSEG */
	      mpgsndra[gsn] = 0xFC00 & (dra + 0x3FF);
					/* Round to next page offset */
	      break;

	    default:                    /* All others */
	      mpgsndra[gsn] = dra;      /* Use byte offset */
	      break;
	  }
      }
    apropSn->as_cbPv = seglen;          /* Save segment length */
    mpsngsn[snMac++] = gsn;             /* Map SEGDEF no. to gsn */
    if(align == ALGNABS) mpsegsa[mpgsnseg[gsn]] = saAbs;
					/* Map seg base to frame number */
    DirtyBp(vbp);                       /* Mark page as changed */
    LEAVE(SegRec2,0xFFFF);              /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  GrpRec2:                                                     *
    *                                                               *
    *  This function processes GRPDEF records on pass 2.            *
    *  See pp. 36-39 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

GrpRec2()
  {
    LNAMETYPE           lnameGroup;     /* Group name index */
    APROPGROUPTYPE      *apropGroup;    /* Pointer to property cell */

    ENTER(GrpRec2);                     /* Debug info */
    lnameGroup = GetIndex(1,lnameMac - 1);
					/* Read in group name index */
    apropGroup = (APROPGROUPTYPE *)
      PropRhteLookup(mplnamerhte[lnameGroup],ATTRGRP,FALSE);
					/* Look up entry in hash table */
    if(grMac >= GRMAX) Fatal("Too many GRPDEFs in module.",0xA2);
					/* Check for table overflow */
    mpgrggr[grMac++] = apropGroup->ag_ggr;
					/* Map GRPDEF to global GRPDEF */
    SkipBytes(cbRec - 1);               /* Skip to checksum byte */
    LEAVE(GrpRec2,0xFFFF);              /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ExtRec2:                                                     *
    *                                                               *
    *  This  function  processes EXTDEF  records  on pass 2.  Note  *
    *  that in pass 2, any undefined externals are errors.          *
    *  See pp. 47-48 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

ExtRec2()
  {
    SBTYPE              sb;             /* External symbol name */
    APROPNAMETYPE       *apropName;     /* Property cell pointer */
    APROPUNDEFTYPE      *apropUndef;    /* Property cell pointer */

    ENTER(ExtRec2);                     /* Debug info */
    while(cbRec > 1)                    /* While not at end of record */
      {
	sb[0] = Gets();                 /* Read in symbol length */
	GetBytes(&sb[1],B2W(sb[0]));    /* Read in the text of the symbol */
	GetIndex(0,0x7FFF);             /* Skip the type index */
	if(extMac >= EXTMAX)
	  Fatal("Too many external symbols in one module.",0xA3);
					/* Check for table overflow */
	apropName = (APROPNAMETYPE *) PropSymLookup(sb,ATTRPNM,FALSE);
					/* Look for symbol among PUBDEFs */
	if(apropName == PROPNIL)        /* If not among PUBDEFs */
	  {
	    fUndefinedExterns = TRUE;   /* There are undefined externals */
	    apropUndef = (APROPUNDEFTYPE *) PropSymLookup(sb,ATTRUND,TRUE);
					/* Look among undefs; add, perhaps */
	    if(!apropUndef->au_rFil) ++cErrors;
	    else apropUndef = (APROPUNDEFTYPE *) PropAdd(vrhte,ATTRUND);
	    apropUndef->au_rFil = vrpropFile;
	    mpextra[extMac] = 0;        /* No known offset */
	    mpextgsn[extMac] = SNNIL;   /* No known SEGDEF */
	  }
	else
	  {
	    mpextra[extMac] = apropName->an_ra;
					/* Save the offset */
	    mpextgsn[extMac] = apropName->an_gsn;
					/* Save the global SEGDEF number */
	  }
	++extMac;                       /* Increment public symbol counter */
      }
    LEAVE(ExtRec2,0xFFFF);              /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ObExpandIteratedData:                                        *
    *                                                               *
    *  This  function  expands  a  LIDATA  record  and moves it to  *
    *  virtual memory.  The  function  returns  a  pointer to  the  *
    *  start of the next iterated  data block (if any).  This is a  *
    *  recursive function.                                          *
    *  See pp. 68-69,63 in "8086 Object Module Formats EPS."        *
    *                                                               *
    ****************************************************************/

LOCAL BYTE              *ObExpandIteratedData(pb,cBlocks)
BYTE                    *pb;            /* Pointer into LIDATA buffer */
WORD                    cBlocks;        /* Current block count subfield */
  {
    WORD                cNextLevelBlocks;
					/* Block count for next level */
    WORD                cRepeat;        /* Repeat count */
    WORD                cbContent;      /* Size of content subfield in bytes */
    BYTE                *pbRet;         /* Recursion return value */

    ENTER(ObExpandIteratedData);        /* Debug info */
    DEBUGVALUE(pb);                     /* Debug info */
    DEBUGVALUE(cBlocks);                /* Debug info */
    DEBUGVALUE(&rgmi[vcbData]);         /* Debug info */
    DEBUGVALUE(vraCur);                 /* Debug info */
    if(!cBlocks)                        /* If block count subfield is zero */
      {
	cbContent = B2W(*pb++);         /* Get size of content subfield */
#if NOT OIAPX286
	MoveToVm(cbContent,pb,AREAFSG + vsegCur,vraCur);
					/* Move data to virtual memory */
	if(pb[vcbData])
	  RecordSegmentReference(vsegCur,vraCur,B2W(pb[vcbData]));
#endif
#if OIAPX286
	MoveToVm(cbContent,pb,AREAFSG+vsegCur,vraCur-mpsegraFirst[vsegCur]);
					/* Move data to virtual memory */
#endif
	vraCur += cbContent;            /* Adjust current offset */
	pb += cbContent;                /* Move ahead in buffer */
      }
    else                                /* Else if non-zero block count */
      {
	while(cBlocks--)                /* While there are blocks to do */
	  {
	    cRepeat = WORDAT(pb,0);     /* Get repeat count */
	    cNextLevelBlocks = WORDAT(pb,2);
					/* Get block count */
	    pb += 4;                    /* Skip over fields */
	    ASSERT(cRepeat != 0);       /* One hopes it won't happen */
	    if(!cRepeat) InvalidObject();
					/* Must have non-zero repeat count */
	    while(cRepeat--) pbRet = ObExpandIteratedData(pb,cNextLevelBlocks);
					/* Recurse to expand record */
	    pb = pbRet;                 /* Skip over expanded block */
	  }
      }
    DEBUGVALUE(pb);                     /* Debug info */
    DEBUGVALUE(rgmi + vcbData + 1);     /* Debug info */
    ASSERT(pb <= rgmi + vcbData + 1);   /* Should be true always */
    if(pb > rgmi + vcbData + 1) InvalidObject();
					/* Length must agree with format */
    LEAVE(ObExpandIteratedData,pb);     /* Debug info */
    return(pb);                         /* Ret ptr to next iterated data blk */
  }

    /****************************************************************
    *                                                               *
    *  DataBlockToVm:                                               *
    *                                                               *
    *  This function moves data  from a LEDATA record  or a LIDATA  *
    *  record into virtual memory.                                  *
    *  See pp. 66-69 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

LOCAL                   DataBlockToVM()
  {
    BYTE                *pb;            /* Pointer to data buffer */

    ENTER(DataBlockToVM);               /* Debug info */
    DEBUGVALUE(vcbData);                /* Debug info */
    if(vrectData == LEDATA)             /* If enumerated data record */
      {
	DEBUGVALUE(AREAFSG + vsegCur);  /* Debug info */
	DEBUGVALUE(vraCur);             /* Debug info */
#if NOT OIAPX286
	MoveToVm(vcbData,rgmi,AREAFSG + vsegCur,vraCur);
					/* Move data to virtual memory */
#endif
#if OIAPX286
	MoveToVm(vcbData,rgmi,AREAFSG+vsegCur,vraCur-mpsegraFirst[vsegCur]);
					/* Move data to virtual memory */
#endif
	vraCur += vcbData;              /* Update current offset */
      }
    else                                /* Else if iterated data record */
      {
	pb = rgmi;                      /* Get address of buffer */
	while((pb = ObExpandIteratedData(pb,1)) < rgmi + vcbData);
					/* Expand and move to VM */
      }
    vrectData = RECTNIL;                /* Reset to nil */
    DEBUGVALUE(vsegCur);                /* Debug info */
    mpsegFlags[vsegCur] |= FNOTEMPTY;   /* Current segment not empty */
    LEAVE(DataBlockToVM,0xFFFF);        /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  LinRec2:                                                     *
    *                                                               *
    *  This function processes LINNUM records on pass 2.            *
    *  See pp. 51-52 in "8086 Object Module Formats EPS."           *
    *                                                               *
    ****************************************************************/

LinRec2()
  {
    SNTYPE              sn;             /* SEGDEF index value */
    SNTYPE              gsn;            /* Global SEGDEF no. */
    SEGTYPE             seg;
    WORD                ln;             /* Line number */
    RATYPE              ra;             /* Offset */
    GRTYPE              gr;             /* Group index */
    APROPTYPE           *aprop;
    AHTETYPE            *ahte;          /* Pointer to hash table entry */

    ENTER(LinRec2);                     /* Debug info */
    gr = GetIndex(0,grMac - 1);         /* Get GRPDEF index */
    ASSERT(gr == 0);                    /* Non-zero GRPDEF ind not supported */
    sn = GetIndex(0,snMac - 1);         /* Get SEGDEF index */
    ASSERT(sn != 0);                    /* Frame number not supported */
    gsn = mpsngsn[sn];                  /* Get global SEGDEF number */
    seg = mpgsnseg[gsn];
    if(gsn != vgsnLineNosPrev)          /* If we weren't doing line numbers */
      {                                 /* for this segment last time */
	if(vgsnLineNosPrev) NEWLINE(bsLst);
					/* Newline */
	NEWLINE(bsLst);                 /* Skip a line */
	fputs("Line numbers for ",bsLst);
					/* Message */
	aprop = (APROPTYPE *) ObGetVa(AREASYMS,mpgsnrprop[gsn],FALSE);
					/* Fetch from virtual memory */
	ASSERT(aprop != PROPNIL);       /* Should never happen! */
	while(aprop->a_attr != ATTRNIL) /* While not at end of prop list */
	  {
	    aprop = (APROPTYPE *) ObGetVa(AREASYMS,aprop->a_next,FALSE);
					/* Move down the list */
	  }
	ahte = (AHTETYPE *) aprop;      /* Recast pointer */
	OutSb(bsLst,ahte->cch);
	NEWLINE(bsLst);                 /* End of line */
	NEWLINE(bsLst);                 /* Skip a line */
	vgsnLineNosPrev = gsn;          /* Save global SEGDEF number */
	vcln = 0;                       /* No entries on line yet */
      }
    while(cbRec > 1)                    /* While not at checksum */
      {
	if(vcln >= 4)                   /* If four entries on this line */
	  {
	    vcln = 0;                   /* Reset counter */
	    NEWLINE(bsLst);             /* Newline */
	  }
	ln = WGets();                   /* Read in line number */
	ra = WGets() + mpgsndra[gsn];   /* Get fixed segment offset */
	fprintf(bsLst,"  %d %04x:%04x",ln,mpsegsa[seg],ra);
	++vcln;                         /* Increment counter */
      }
    LEAVE(LinRec2,0xFFFF);              /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ProcP2:                                                      *
    *                                                               *
    *  This  function  controls the processing of  object  modules  *
    *  during pass 2.                                               *
    *                                                               *
    ****************************************************************/

ProcP2()
  {
#if LIBXENIX
    LFATYPE             lfa;            /* Seek value */
#endif

    ENTER(ProcessP2);                   /* Debug info */
    snMac = 1;                          /* Initialize counter */
    grMac = 1;                          /* Initialize counter */
    extMac = 1;                         /* Initialize counter */
    lnameMac = 1;                       /* Initialize counter */
    vrectData = RECTNIL;                /* Initialize record type variable */
#if LIBXENIX
    lfa = ftell(bsInput);               /* Save initial file position */
    cbRec = WGets();                    /* Read a word */
    if(cbRec == X_MAGIC)                /* If magic number found */
      {
	cbRec = 30;                     /* Size of x.out header minus 2 */
	SkipBytes(26);                  /* Skip to cpu specification */
	if(Gets() != XC_8086) InvalidObject();
					/* Bad if not 8086 */
	if(Gets() != XR_R86REL + XR_S86REL) InvalidObject();
	if(WGets() != XE_V2) InvalidObject();
      }
    else fseek(bsInput,lfa,0);          /* Else return to start */
#endif
    for(;;)                             /* Loop to process object module */
      {
	rect = Gets();                  /* Read record type */
	cbRec = WGets();                /* Read record length */
	DEBUGVALUE(rect);               /* Debug info */
	DEBUGVALUE(cbRec);              /* Debug info */
	if(!cbRec || B2W(rect) < 0x6E || B2W(rect) > 0xAA || (rect & 1)) break;
					/* Break if invalid object */
	if(rect == FIXUPP) RelocRec();  /* If FIXUPP, perform relocation */
	else                            /* Else if not fixup record */
	  {
	    if(vrectData != RECTNIL) DataBlockToVM();
					/* Move data to virtual memory */
	    switch(rect)                /* Switch on record type */
	      {
		case SEGDEF:
		  SegRec2();
		  break;

		case THEADR:
		  ModRc1();
		  break;

		case COMENT:
		  ComRc1();
		  break;

		case LHEADR:
		  ModRc1();
		  break;

		case GRPDEF:
		  GrpRec2();
		  break;

		case EXTDEF:
		  ExtRec2();
		  break;

		case LNAMES:
		  LNmRc1();
		  break;

		case LINNUM:
		  if(vfLineNos) LinRec2();
					/* Print line numbers if requested */
		  else SkipBytes(cbRec - 1);
					/* Else skip to checksum */
		  break;

		case LEDATA:
		case LIDATA:
		  DataRec();
		  break;

		case MODEND:            /* End of module */
		  EndRec();             /* Process MODEND */
		  LEAVE(ProcessP2,0xFFFF);
					/* Debug info */
		  return;               /* And return */

		default:
		  SkipBytes(cbRec - 1); /* Skip to checksum byte */
		  break;
	      }
	  }
	DEBUGVALUE(cbRec);              /* Debug info */
	if(cbRec != 1) break;           /* If record length bad */
	Gets();                         /* Eat the checksum byte */
      }
    InvalidObject();                    /* Invalid module */
  }
