    /****************************************************************
    *                                                               *
    *                         NEWTRN.C                              *
    *                                                               *
    *  Main function of the linker.                                 *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types, constants, macros */
#include                "bndtrn.h"      /* Types and constants */
#include                "lnkio.h"       /* Linker I/O definitions */

#if OSMSDOS OR OSPCDOS OR OSCPM86
extern BYTE             SCPCurrentDisk();
					/* Determine current disk drive */
#if VERSION2
extern FTYPE            FChangeable();  /* True if drive is floppy */
#endif
#endif
extern                  LibrarySearch();/* Search libraries to find undefs */
extern char             *ObGetVa();     /* Fetch from virtual memory */
extern                  FinishVm();     /* Close and delete VM file */
extern                  ProcP1();       /* Perform pass 1 processing on mod */
extern                  PrintMap();     /* Print link map */
extern                  AssignAddresses();
					/* Assign final addrs to segments */
extern                  ProcP2();       /* Perform pass 2 processing on mod */
extern                  OutRunFile();   /* Write the executable file */
extern                  EnSyms();       /* Apply proc to syms w/ given attr */
extern                  ParseCmdLine(); /* Parse command line! */

extern WORD             cbRgRleMac;     /* Count of bytes in reloc buffer */
extern RECTTYPE         vrectData;      /* Type of data record (if any) */
extern FTYPE            fLstFileOpen;   /* Map-file-open flag */
extern FTYPE            vfDSAlloc;      /* Perform-DS-allocation flag */
extern BYTE             chRunFile;      /* Disk drive for run file */
extern WORD             cparMaxAlloc;
extern BSTYPE           bsInput;        /* File handle for input file */
extern BSTYPE           bsLst;          /* File handle for map file */
extern BSTYPE           bsRunfile;      /* File handle for executable file */
extern char             *vob;           /* Pointer to current virtual object */
extern FTYPE            fUndefinedExterns;
					/* Undefined externals flag */
extern HTETYPE          vhte;           /* Pointer to current hash tab entry */
extern WORD             cbStack;        /* Size of stack in bytes */
extern RECTTYPE         rect;           /* Type of current object record */
extern SNTYPE           vgsnLineNosPrev;/* Global SEGDEF no. for which line */
					/* numbers were last emitted */
extern FTYPE            fLibPass;
WORD                    ifhLast;
extern RBTYPE           vrpropFile;
extern RBTYPE           vrhteFile;
extern FTYPE            vfPass1;        /* Pass 1 flag */
extern WORD             cErrors;        /* Count of non-fatal errors */
extern RBTYPE           rprop1stFile;
FTYPE                   fDrivePass;
FTYPE                   fRunFileOpen;   /* Executable-file-open flag */
extern FTYPE            fPauseRun;
extern SEGTYPE          mpgsnseg[];     /* f(glob SEGDEF) = seg order */
extern RATYPE           mpgsndra[];     /* f(glob SEGDEF) = dra */
extern RBTYPE           mpgsnrprop[];   /* f(glob SEGDEF) = prop list addr */
extern FTYPE            mpsegFlags[];
extern WORD             mpifhAlgnF[];   /* f(lib no.) = internal page size */
#if NOT OIAPX286
extern WORD             mpiovcRle[];    /* Count of long reloc.s per overlay */
#endif
extern RBTYPE           mpiovrhte[];    /* f(overlay no.) = hash tab addr */
#if OSMSDOS OR OSPCDOS OR OSCPM86
BYTE                    chListFile;     /* Disk drive for map file */
#endif
extern FTYPE            fScrClosed;     /* Screen-closed flag */
extern char             *lnknam;        /* Linker name (at runtime) */

#if DEBUG
    /****************************************************************
    *                                                               *
    *  DmpBlk:                                                      *
    *                                                               *
    *  Dump a specified block of  bytes of a specified length to a  *
    *  specified file.                                              *
    *                                                               *
    ****************************************************************/

DmpBlk(fd,pb,cb)
BSTYPE                  fd;             /* File descriptor */
BYTE                    *pb;            /* Byte pointer */
WORD                    cb;             /* No. of bytes to dump */
  {
    BYTE                ascii[16];      /* Ascii value buffer */
    WORD                i;              /* Counter */
    WORD                j;              /* Counter */

    i = 0;                              /* Initialize */
    j = 0;                              /* Initialize */
    while(cb--)                         /* While characters remain */
      {
	if(!j)                          /* If at start of line */
	  {
	    fprintf(fd,"%03x0: ",i);    /* Offset */
	  }
	if(B2W(*pb) >= 0x20 && B2W(*pb) < 0x7F) ascii[j] = *pb;
					/* If printable, print */
	else ascii[j] = '.';            /* Else print dot */
	fprintf(fd," %02x",B2W(*pb++)); /* Print byte */
					/* Print byte in hex */
	if(++j == 16)                   /* If end of line */
	  {
	    j = 0;                      /* Reset j */
	    ++i;                        /* Increment i */
	    OutByte(fd,' ');            /* Space */
	    fwrite(ascii,1,16,fd);      /* Write ascii values */
	    NEWLINE(fd);                /* End of line */
	  }
      }
    if(j)                               /* If in middle of line */
      {
	for(i = j; i < 16; ++i) fputs("   ",fd);
					/* Write blank fields */
	OutByte(fd,' ');                /* Space */
	fwrite(ascii,1,j,fd);           /* Write ascii values */
	NEWLINE(fd);                    /* End of line */
      }
    fflush(fd);                         /* Force to file */
  }
#endif

    /****************************************************************
    *                                                               *
    *  CleanUp:                                                     *
    *                                                               *
    *  This function cleans up after the rest of the linker.        *
    *                                                               *
    ****************************************************************/

CleanUp()
  {
    ENTER(CleanUp);                     /* Debug info */
    if(fRunFileOpen) fclose(bsRunfile); /* If run file open, close it */
    if(vgsnLineNosPrev) NEWLINE(bsLst); /* Write newline */
    if(fLstFileOpen) fclose(bsLst);     /* Close map file if open */
#if CMDMSDOS
    if(cErrors)                         /* If there were non-fatal errors */
      {
	NEWLINE(stderr);                /* Newline */
	fprintf(stderr,"There w%s %d error%s detected.",
	  cErrors == 1? "as": "ere",cErrors,cErrors == 1? "": "s");
	NEWLINE(stderr);                /* Newline */
	fflush(stderr);                 /* Flush buffer */
      }
#endif
    FinishVm();                         /* Close and delete VM file if any */
    LEAVE(CleanUp,0xFFFF);              /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ErrMes:                                                      *
    *                                                               *
    *  This function closes the  listing file and  outputs a fatal  *
    *  error message to the screen.                                 *
    *                                                               *
    ****************************************************************/

LOCAL                   ErrMes(pasciz)
char                    *pasciz;        /* Null-terminated message string */
  {
#if CMDXENIX
    fprintf(stderr,"%s: ",lnknam);      /* Linker name */
#endif
    fputs(pasciz,stderr);               /* Error message */
    NEWLINE(stderr);                    /* Newline */
    fflush(stderr);                     /* Force message to screen */
  }

    /****************************************************************
    *                                                               *
    *  OutError:                                                    *
    *                                                               *
    *  This function outputs a non-fatal error message.             *
    *                                                               *
    ****************************************************************/

OutError(bs,pasciz)
BSTYPE                  bs;             /* File handle */
char                    *pasciz;        /* Message */
  {
    ENTER(OutError);                    /* Debug info */
    if(fLstFileOpen && bs == bsLst && vgsnLineNosPrev)
      {                                 /* If we've listed line numbers */
	NEWLINE(bs);                    /* Newline */
	vgsnLineNosPrev = 0;            /* Reset */
      }
    fputs(pasciz,bs);                   /* Write message to screen */
    NEWLINE(bs);                        /* Newline */
    fflush(bs);                         /* Force message out */
    LEAVE(OutError,0xFFFF);             /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ErrEnd:                                                      *
    *                                                               *
    *  This function terminates  execution in the event of a fatal  *
    *  error.                                                       *
    *                                                               *
    ****************************************************************/

LOCAL                   ErrEnd(erc)
WORD                    erc;            /* Error code */
  {
    LFATYPE             lfa;            /* Logical file address */

    if(fDrivePass)
      {
	fputs("Input File: ",stderr);   /* Message */
	OutFileCur(stderr);             /* Write name of current input file */
	lfa = ftell(bsInput);           /* Get file offset */
	fprintf(stderr," pos: %lx Record type: %02x",lfa,B2W(rect));
	NEWLINE(stderr);                /* Newline */
	fflush(stderr);                 /* Force message to screen */
      }
    if(fLstFileOpen) fclose(bsLst);     /* Close map file (if any) */
    KillRunfile();                      /* Close and delete run file if any */
    FinishVm();                         /* Close and delete VM file if any */
    exit(erc);                          /* Exit */
  }

    /****************************************************************
    *                                                               *
    *  OutFileCur:                                                  *
    *                                                               *
    *  This function outputs the name of the current input file.    *
    *                                                               *
    ****************************************************************/

OutFileCur(bs)
BSTYPE                  bs;             /* File handle */
  {
    APROPFILETYPE       *apropFile;     /* Pointer to file property cell */
    AHTETYPE            *ahte;          /* Pointer to hash table entry */
    RBTYPE              rhteModule;     /* Pointer to module hash tab entry */

    ENTER(OutFileCur);                  /* Debug info */
    DEBUGVALUE(vrpropFile);             /* Debug info */
    apropFile = (APROPFILETYPE *) ObGetVa(AREASYMS,vrpropFile,FALSE);
					/* Fetch from virtual memory */
    rhteModule = apropFile->af_rMod;    /* Get pointer to module entry */
    while(apropFile->af_attr != ATTRNIL)
      {                                 /* While not at end of list */
	DEBUGVALUE(apropFile->af_next); /* Debug info */
	apropFile = (APROPFILETYPE *)
	  ObGetVa(AREASYMS,apropFile->af_next,FALSE);
					/* Move down the list */
      }
    ahte = (AHTETYPE *) apropFile;      /* Recast pointer */
    OutSb(bs,ahte->cch);                /* Output file name */
    if(rhteModule)                      /* If we have a module name */
      {
	OutByte(bs,'(');                /* Open parens */
	ahte = (AHTETYPE *) ObGetVa(AREASYMS,rhteModule,FALSE);
					/* Fetch hash tab ent from VM */
	OutSb(bs,ahte->cch);            /* Output module name */
	OutByte(bs,')');                /* Close parens */
      }
    LEAVE(OutFileCur,0xFFFF);           /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  Fatal:                                                       *
    *                                                               *
    *  This function prints an  error message and does its best to  *
    *  tell where the error occurred.                               *
    *                                                               *
    ****************************************************************/

Fatal(pasciz,erc)
char                    *pasciz;        /* Error message */
WORD                    erc;            /* Error code */
  {
    ErrMes(pasciz);                     /* Write error message */
    ErrEnd(erc);                        /* Tell where error was */
  }

    /****************************************************************
    *                                                               *
    *  FatalFile:                                                   *
    *                                                               *
    *  This function is called when an input file cannot be opened  *
    *  or is not of the proper format.                              *
    *                                                               *
    ****************************************************************/

FatalFile(psd,erc)
SDTYPE                  *psd;           /* Ptr to struct with name info */
WORD                    erc;            /* Error code */
  {
#if CMDXENIX
    fprintf(stderr,"%s: ",lnknam);      /* Linker name */
#endif
    fwrite(psd->pb,1,psd->cb,stderr);   /* Write name to console */
    if(erc) fputs(": cannot open file",stderr);
					/* Message */
    else fputs(": invalid format file",stderr);
					/* Message */
    NEWLINE(stderr);                    /* Newline */
    fflush(stderr);                     /* Force message to screen */
    fDrivePass = FALSE;
    ErrEnd(erc);                        /* Exit abnormally */
  }

    /****************************************************************
    *                                                               *
    *  ErrHlt:                                                      *
    *                                                               *
    *  This function is  called to halt execution  when the linker  *
    *  is unable to handle a valid link request.                    *
    *                                                               *
    ****************************************************************/

ErrHlt(erc)
WORD                    erc;            /* Error code */
  {
    ErrMes("An internal failure has occured.");
					/* Message */
    ErrEnd(erc);                        /* Tell where */
  }

    /****************************************************************
    *                                                               *
    *  DrivePass:                                                   *
    *                                                               *
    *  This  function  applies  either  the  pass 1 or  the pass 2  *
    *  object module  processor  to  all  the objects being linked  *
    *  together.                                                    *
    *                                                               *
    ****************************************************************/

DrivePass(pProcessPass)
PFUN                    pProcessPass;   /* Function pointer */
  {
    AHTETYPE            *ahte;          /* Pointer to hash table entry */
    APROPFILETYPE       *apropFile;     /* Pointer to file entry */
    WORD                ifh;            /* File handle index */
#if LIBMSDOS
    WORD                pnFirst;        /* First virtual page number */
#endif
    RBTYPE              rbFileNext;     /* Ptr to prop list of next file */
    LFATYPE             lfa;            /* File offset */
    BSTYPE              bsTmp;          /* Place to save file handle */
    FTYPE               fLibraryFile;   /* True if input from library */
    BYTE                rgbTmp[4];
    BYTE                b;              /* A byte */
    SBTYPE              sbInput;        /* Input file name */

    ENTER(DrivePass);                   /* Debug info */
    fDrivePass = TRUE;                  /* Executing DrivePass */
    rbFileNext = rprop1stFile;          /* Next file to look at is first */
    while(rbFileNext)                   /* Loop to process objects */
      {
	vrpropFile = rbFileNext;        /* Make next file the current file */
	apropFile = (APROPFILETYPE *) ObGetVa(AREASYMS,vrpropFile,FALSE);
					/* Fetch table entry from VM */
	rbFileNext = apropFile->af_FNxt;/* Get pointer to next file */
	ifh = apropFile->af_ifh;        /* Get the file handle index */
	fLibraryFile = (ifh != FHNIL);  /* Set library flag */
#if LIBMSDOS
	pnFirst = apropFile->af_pn1;    /* Get page no. of 1st page */
#endif
#if LIBXENIX
	lfa = apropFile->af_lfa;        /* Get file offset */
#endif
	/* "Get hte (name) of file" */
	while(apropFile->af_attr != ATTRNIL)
	  {                             /* While haven't found nil attr */
	    vrhteFile = apropFile->af_next;
					/* Try next entry in list */
	    apropFile = (APROPFILETYPE *) ObGetVa(AREASYMS,vrhteFile,FALSE);
					/* Fetch it from VM */
	  }
	DEBUGVALUE(vrhteFile);          /* Debug info */
	vhte = (HTETYPE) apropFile;     /* Save pointer to hash tab entry */
	ahte = (AHTETYPE *) vhte;       /* Overlay struct */
#if DEBUG                               /* If debugging on */
	fputs("  File ",stderr);        /* Message */
	OutSb(stderr,ahte->cch);        /* File name */
	if(fLibraryFile) fputs(" (library)",stderr);
					/* Message */
	NEWLINE(stderr);                /* Newline */
	fflush(stderr);                 /* Force message to screen */
#endif                                  /* End debugging code */
#if OSMSDOS OR OSPCDOS OR OSCPM86
#if NOT VERSION2
	/* "Check for control-c" */
	write(STDOUT," \010",2);        /* Write space-backspace */
#endif
#endif
	if(!fLibraryFile || ifhLast != ifh)
	  {                             /* If new file to open */
	    if(!fLibPass && ifhLast != FHNIL) fclose(bsInput);
					/* Close previous lib. on pass two */
#if OSMSDOS OR OSPCDOS OR OSCPM86
	    for(;;)                     /* Loop to get input file and */
	      {                         /* allow user to change diskette */
#endif
		Move(B2W(ahte->cch[0]) + 1,ahte->cch,sbInput);
					/* Copy name to buffer */
		sbInput[B2W(ahte->cch[0]) + 1] = '\0';
					/* Null-terminate file name */
		if((bsInput = fopen(&sbInput[1],RDBIN)) != NULL)
		  {                     /* If no error opening input file */
		    DEBUGVALUE(bsInput);/* Debug info */
		    DEBUGVALUE(FileNo(bsInput));
					/* Debug info */
		    SETRAW(bsInput);    /* Binary file */
#if OSMSDOS OR OSPCDOS OR OSCPM86
		    break;              /* Exit loop */
#endif
		  }
#if OSTOPS20 OR OSXENIX
		else                    /* Else if error occurred */
		  {
#endif
#if OSMSDOS OR OSPCDOS OR OSCPM86
		    if(ahte->cch[2] == ':') b = ahte->cch[1] - 'A';
					/* If disk specified, grab it */
		    else b = SCPCurrentDisk();
					/* Else use current drive */
		    bsTmp = bsLst;      /* Save map file handle */
#endif
		    bsLst = stderr;     /* Force message to screen */
#if CMDXENIX
		    fprintf(stderr,"%s: ",lnknam);
					/* Linker name */
#endif
		    fputs("cannot find file ",stderr);
					/* Message */
		    OutSb(stderr,ahte->cch);
					/* Output file name */
		    NEWLINE(stderr);    /* Newline */
		    fflush(stderr);     /* Force message to screen */
#if OSMSDOS OR OSPCDOS OR OSCPM86
		    /* "If we are changeing the listfile device or
		    *  the VM.TMP device or if the device is not
		    *  changeable, then exit."
		    */
		    if((b == chListFile && bsTmp != STDOUT &&
		      bsTmp != FHNUL) || (!fScrClosed && b == SCPCurrentDisk())
		      || !FChangeable(b))
		      {
#endif
			exit(1);        /* Exit */
#if OSMSDOS OR OSPCDOS OR OSCPM86
		      }
#endif
#if OSTOPS20 OR OSXENIX
		  }
#endif
#if OSMSDOS OR OSPCDOS OR OSCPM86
		for(;;)                 /* Loop to prompt for new drive */
		  {
		    fputs("change diskette and hit <ENTER>",stderr);
					/* Prompt message */
		    fflush(stderr);     /* Force prompt to screen */
		    if(read(0,rgbTmp,4) == 1) break;
					/* Read response */
		  }
		bsLst = bsTmp;          /* Restore map file handle */
	      }
#endif
	  }
#if LIBMSDOS
	if(fLibraryFile) lfa = (LFATYPE) pnFirst << mpifhAlgnF[ifh];
					/* If library, find file offset */
	else lfa = 0L;                  /* Else start at the beginning */
#endif
	fseek(bsInput,lfa,0);           /* Seek to desired offset */
	(*pProcessPass)();              /* Call ProcP1 or ProcP2 */
	ifhLast = ifh;                  /* Save this file handle */
	if(!fLibraryFile)               /* If not a library */
	  {
	    fclose(bsInput);            /* Close input file */
	  }
      }
    if(ifh != FHNIL && !fLibPass) fclose(bsInput);
					/* Close last library on pass two */
    fDrivePass = FALSE;                 /* No longer executing DrivePass */
    LEAVE(DrivePass,0xFFFF);            /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  PrintAnUndef:                                                *
    *                                                               *
    *  This  function will print  the name of an  undefined symbol  *
    *  and the name(s) of the module(s) in which it is referenced.  *
    *  This routine is passed as an argument to EnSyms().           *
    *                                                               *
    ****************************************************************/

PrintAnUndef(hte,prop,rhte,rprop,fNewHte)
AHTETYPE                *hte;           /* Pointer to hash table entry */
APROPUNDEFTYPE          *prop;          /* Pointer to undef prop cell */
RBTYPE                  rprop;          /* Virt addr of prop cell */
RBTYPE                  rhte;           /* Virt addr of hash tab ent */
FTYPE                   fNewHte;        /* True if name has been written */
  {
    if(!prop->au_rFil) return;          /* "sbVer etc..." */
    if(fNewHte)                         /* If name hasn't been written yet */
      {
	NEWLINE(bsLst);                 /* Newline */
	OutSb(bsLst,hte->cch);          /* Write symbol name */
	fputs(" in file(s): ",bsLst);   /* Message */
	NEWLINE(bsLst);                 /* Newline */
      }
    OutByte(bsLst,' ');                 /* Write a space */
    vrpropFile = prop->au_rFil;         /* Set ptr to ent for refrncing file */
    OutFileCur(bsLst);                  /* Output file name */
  }

    /****************************************************************
    *                                                               *
    *  MainFn:                                                      *
    *                                                               *
    *  Not really the main function, but more or less the main  *
    *  function.                                                    *
    *                                                               *
    ****************************************************************/

MainFn()
  {
    register WORD       i;              /* Index variable */
    BSTYPE              bsTmp;          /* File handle saver */
    BYTE                junk[2];        /* Junk */

#if CMDMSDOS
    NEWLINE(stdout);                    /* Skip a line */
#if OSPCDOS
    fputs("   IBM 8086 Object Linker VX.XX",stdout);
#else
    fputs("   Microsoft 8086 Object Linker V2.10",stdout);
#endif
    NEWLINE(stdout);                    /* Newline */
#if OSPCDOS
    fputs("(C) Copyright 1983 by IBM Corp.",stdout);
#else
    fputs("(C) Copyright 1983 by Microsoft Inc.",stdout);
#endif
    NEWLINE(stdout);                    /* Newline */
#if OXOUT
    fputs("   (Produces XENIX 8086 x.outs)",stdout);
					/* Message */
    NEWLINE(stdout);                    /* Newline */
#endif
#if OIAPX286
    fputs("   (Produces XENIX 80286 x.outs)",stdout);
					/* Message */
    NEWLINE(stdout);                    /* Newline */
#endif
#if LIBXENIX
    fputs("   (Takes XENIX 8086 libraries)",stdout);
					/* Message */
    NEWLINE(stdout);                    /* Newline */
#endif
    NEWLINE(stdout);                    /* Skip a line */
    fflush(stdout);                     /* Force banner to screen */
#endif
    ParseCmdLine();                     /* Parse the command line */

    for(i = 0; i < SEGMAX; ++i)         /* Loop to initialize tables */
      {
	mpgsnseg[i] = SEGNIL;
	mpgsndra[i] = 0;
	mpgsnrprop[i] = 0;
	mpsegFlags[i] = FALSE;
      }
    if(fLstFileOpen)                    /* If we have a list file */
      {
#if OSPCDOS
	if(bsLst != stdout)             /* If map not going to terminal */
	  {
#endif
	    if(vfDSAlloc) fputs("DS Allocation  ",bsLst);
					/* Tell if DS-allocating */
	    if(!cparMaxAlloc) fputs("Loading High  ",bsLst);
					/* Tell if loading high */
	    if(cbStack)                 /* If stack size specified */
	      {
		fprintf(bsLst,"Stack Allocation = %d bytes",cbStack);
					/* Print stack size */
	      }
	    NEWLINE(bsLst);             /* Newline */
#if OSPCDOS
	  }
#endif
      }
    ifhLast = FHNIL;                    /* No files looked at yet */
    DEBUGMSG("**** PASS ONE ****");     /* Message */
    DrivePass(ProcP1);                  /* Make pass 1 */
    DEBUGMSG("**** LIBRARY SEARCH ****");
					/* Message */
    LibrarySearch();                    /* Search libraries */
    vrectData = RECTNIL;                /* Not looking at a data record */
    DEBUGMSG("**** ASSIGN ADDRESSES ****");
					/* Message */
    AssignAddresses();                  /* Assign addresses to segments */
    if(fLstFileOpen)                    /* If list file open */
      {
	DEBUGMSG("**** PRINT MAP ****");/* Message */
	PrintMap();                     /* Print load map */
      }
    vfPass1 = FALSE;                    /* No longer pass 1 */
    for(i = 0; i < IOVMAX; ++i)         /* Initialize overlay tables */
      {
#if NOT OIAPX286
	mpiovcRle[i] = 0;               /* Init. long reloc. counts */
#endif
	mpiovrhte[i] = RHTENIL;
      }
    ifhLast = FHNIL;                    /* No files examined on pass 2 yet */
    DEBUGMSG("**** PASS TWO ****");     /* Message */
    DrivePass(ProcP2);                  /* Make pass 2 */
#if OSMSDOS OR OSPCDOS OR OSCPM86
    if(chRunFile >= 'a' && chRunFile <= 'z') chRunFile += 'A' - 'a';
					/* Make drive letter upper case */
    if(FChangeable(chRunFile - 'A') && fPauseRun)
      {
	fputs("About to generate .EXE file",stdout);
					/* Message */
	NEWLINE(stdout);                /* Newline */
	fputs("Change diskettes and hit <ENTER>",stdout);
	fflush(stdout);                 /* Force message to screen */
	read(0,junk,2);                 /* Read newline */
      }
#endif
    DEBUGMSG("**** OUTPUT RUN FILE ****");
					/* Message */
    EnSyms(OutRunFile,ATTROVL);         /* Output executable file */
    bsTmp = bsLst;                      /* The old hack again */
    if(fUndefinedExterns)               /* If we have unresolved references */
      {
	bsLst = stdout;                 /* Force message to screen */
	for(;;)                         /* Loop to list undefined syms */
	  {
	    NEWLINE(bsLst);             /* Newline */
#if CMDXENIX
	    fprintf(bsLst,"%s: ",lnknam);
					/* Linker name */
#endif
	    fputs("unresolved externals:",bsLst);
					/* Message */
	    NEWLINE(bsLst);             /* Newline */
	    NEWLINE(bsLst);             /* Skip a line */
	    EnSyms(PrintAnUndef,ATTRUND);
					/* Print undefined symbols */
	    NEWLINE(bsLst);             /* Newline */
	    fflush(bsLst);              /* Flush buffer */
	    if(!fLstFileOpen || bsTmp == stdout) break;
					/* Stop if no list file, or list is
					*  std. out., or 2nd time thru loop
					*/
	    bsLst = bsTmp;              /* Restore map file descriptor */
	    bsTmp = stdout;             /* Insure loop exit */
	  }
      }
    CleanUp();                          /* Mop up after itself */
    exit(cErrors? 1: (fUndefinedExterns? 1: 0));
  }
