    /****************************************************************
    *                                                               *
    *                           NEWUTL.C                            *
    *                                                               *
    *  Linker utilities.                                            *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types, constants */
#include                "bndtrn.h"      /* More types and constants */
#include                "lnkio.h"       /* Linker I/O definitions */

extern                  InvalidObject();/* Error exit procedure */

extern WORD             cbRec;          /* Count of bytes left in record */
extern BSTYPE           bsInput;        /* Input file handle */

    /****************************************************************
    *                                                               *
    *  InvalidObject:                                               *
    *                                                               *
    *  This function calls  Fatal()  with  the  message,  "Invalid  *
    *  object module."   It  does  not  return a meaningful value.  *
    *                                                               *
    ****************************************************************/

InvalidObject()
  {
    Fatal("Invalid object module.",0x90);
					/* This is very bogus! */
  }

    /****************************************************************
    *                                                               *
    *  OutWord:                                                     *
    *                                                               *
    *  Write a 16-bit integer to the specified file.  The low byte  *
    *  precedes the high byte.  Return non-zero on error.           *
    *                                                               *
    ****************************************************************/

WORD                    OutWord(f,w)
FILE                    *f;
register WORD           w;
  {
    if(putc((BYTE)(w & ~(~0 << BYTELN)),f) == EOF) return(EOF);
					/* Write low-order byte */
    return(putc((BYTE)((w >> BYTELN) & ~(~0 << BYTELN)),f));
					/* Write high-order byte */
  }

    /****************************************************************
    *                                                               *
    *  Gets:                                                        *
    *                                                               *
    *  Read a byte of input and return it.                          *
    *                                                               *
    ****************************************************************/

BYTE                    Gets()
  {
    register WORD       b;              /* A byte of input */

    if((b = InByte(bsInput)) == -1) InvalidObject();
					/* Read a byte */
    --cbRec;                            /* One less byte in record */
    return((BYTE) b);                   /* Return the byte */
  }

    /****************************************************************
    *                                                               *
    *  GetBytes:                                                    *
    *                                                               *
    *  Read n bytes from input.                                     *
    *                                                               *
    ****************************************************************/

GetBytes(pb,n)
BYTE                    *pb;            /* Pointer to buffer */
register WORD           n;              /* Number of bytes to read in */
  {
    fread(pb,1,n,bsInput);              /* Ask for n bytes */
    if(ferror(bsInput)) InvalidObject();/* Check for error */
    cbRec -= n;                         /* Update byte count */
  }

    /****************************************************************
    *                                                               *
    *  SkipBytes:                                                   *
    *                                                               *
    *  Skip n bytes of input.                                       *
    *                                                               *
    ****************************************************************/

SkipBytes(n)
register WORD           n;              /* Number of bytes to read in */
  {
    if(fseek(bsInput,(long) n,1)) InvalidObject();
					/* Skip n bytes */
    cbRec -= n;                         /* Update count of bytes remaining */
  }

    /****************************************************************
    *                                                               *
    *  WGets:                                                       *
    *                                                               *
    *  Read a word of input.                                        *
    *                                                               *
    ****************************************************************/

WORD                    WGets()
  {
    register WORD       w;              /* Word of input */

    w = B2W(Gets());                    /* Get low-order byte */
    return(w | (B2W(Gets()) << BYTELN));/* Return word */
  }

    /****************************************************************
    *                                                               *
    *  GetIndex:                                                    *
    *                                                               *
    *  This function  reads in  a variable-length index field from  *
    *  the input file.  It takes as its  arguments two word values  *
    *  which  represent  the minimum and maximum  allowable values  *
    *  the index.  The function returns the value of the index.     *
    *  See p. 12 in "8086 Object Module Formats EPS."               *
    *                                                               *
    ****************************************************************/

WORD                    GetIndex(min,max)
WORD                    min;            /* Minimum permissible value */
WORD                    max;            /* Maximum permissible value */
  {
    register WORD       index;          /* The index */

    if((index = B2W(Gets())) & 0200)
      index = ((index & 0177) << BYTELN) | B2W(Gets());
					/* Read in the index */
    if(index < min || index > max) InvalidObject();
					/* Check for invalid index */
    return(index);                      /* Return a good value */
  }

Setb(bval,pbuf,n)                       /* Set n bytes to given value */
BYTE                    bval;           /* Byte value to fill with */
register BYTE           *pbuf;          /* Address of buffer */
register WORD           n;              /* Number of bytes to set */
  {
    while(n--) *pbuf++ = bval;          /* Do it */
  }

#if CPUOTHER
    /****************************************************************
    *                                                               *
    *  Move:                                                        *
    *                                                               *
    *  This function takes as its  arguments two byte pointers and  *
    *  a WORD n.  It copies n bytes from the location specified by  *
    *  the first pointer to the  location specified by the second.  *
    *  The function does not return a meaningful value.             *
    *                                                               *
    ****************************************************************/

Move(n,s1,s2)                           /* Move byte string */
WORD                    n;              /* Number of bytes to move */
register BYTE           *s1;            /* Source string */
register BYTE           *s2;            /* Destination string */
  {
    while(n--) *s2++ = *s1++;           /* Copy the bytes */
  }
#endif

Setw(val,addr,n)                        /* Initialize word array */
WORD                    val;            /* Word value */
register WORD           *addr;          /* Address of array */
register WORD           n;              /* Length of array */
  {
    register WORD       *pw;            /* Pointer to words */

    pw = addr;                          /* Load register with address */
    while(n--) *pw++ = val;             /* Set the bytes */
  }

WORD                    rol(x,n,size)   /* Rotate left */
register WORD           x;              /* WORD to rotate */
register WORD           n;              /* Bits to rotate */
WORD                    size;           /* Size of word in bits */
  {
    return(((x << n) | ((x >> (size - n)) & ~(~0 << n))) & ~(~0 << size));
  }

WORD                    ror(x,n,size)   /* Rotate right */
WORD                    x;              /* WORD to rotate */
register WORD           n;              /* Bits to rotate */
register WORD           size;           /* Size of word in bits */
  {
    return(((x << (size - n)) | ((x >> n) & ~(~0 << (size - n))))
      & ~(~0 << size));
  }

#if CLATTICE                            /* Lattice C has no fread,fwrite */
int                     fread(pb,cb,co,f)
char                    *pb;            /* Byte pointer */
int                     cb;             /* Size in bytes of objects */
int                     co;             /* Number of objects to read in */
FILE                    *f;             /* Stream to read from */
  {
    int                 i;              /* Index */
    int                 j;              /* Index */

    for(i = 0; i < co; ++i)             /* Loop to read objects */
      {
	for(j = 0; j < cb; ++j)         /* Loop to read each object */
	  {
	    *pb++ = (char) getc(f);     /* Read a char */
	  }
      }
    return(ferror(f)? 0: co);           /* Return 0 or co */
  }

int                     fwrite(pb,cb,co,f)
char                    *pb;            /* Byte pointer */
int                     cb;             /* Size in bytes of objects */
int                     co;             /* Number of objects to write out */
FILE                    *f;             /* Stream to read from */
  {
    int                 i;              /* Index */
    int                 j;              /* Index */

    for(i = 0; i < co; ++i)             /* Loop to read objects */
      {
	for(j = 0; j < cb; ++j)         /* Loop to read each object */
	  {
	    putc(*pb++,f);              /* Write a byte */
	  }
      }
    return(ferror(f)? 0: co);           /* Return 0 or co */
  }
#endif
