    /****************************************************************
    *                                                               *
    *                           BNDVRT.C                            *
    *                                                               *
    *  M. Zbikowski at Microsoft for interface to SCPDOS.           *
    *                                                               *
    *  P. Stewart at Microsoft for conversion from PL/M to C.       *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types and constants */
#include                "bndtrn.h"      /* Types and constants */
#include                "bndvrt.h"
#include                "lnkio.h"       /* Linker I/O definitions */

#if CDEC20                              /* If DEC20 version */
#define cMoves12        l00000          /* Only 6 sig. char.s on DEC20 */
#define cMoves13        l00001          /* Only 6 sig. char.s on DEC20 */
#define cMoves23        l00002          /* Only 6 sig. char.s on DEC20 */
#define cMoves21        l00003          /* Only 6 sig. char.s on DEC20 */
#define cMoves31        l00004          /* Only 6 sig. char.s on DEC20 */
#define cMoves01        l00005          /* Only 6 sig. char.s on DEC20 */
#endif                                  /* End DEC20 definitions */

extern long             lseek();        /* File positioner */
extern                  ErrHlt();       /* Internal failure error function */
extern                  Fatal();        /* Fatal error function */
#if NOT CDEC20
extern char             *malloc();      /* Storage allocator */
#endif
#if OSXENIX OR OSTOPS20
extern char             *mktemp();      /* Temporary file name generator */
#endif
#if OSMSDOS OR OSPCDOS OR OSCPM86
extern BYTE             SCPCurrentDisk();
					/* Determine current default drive */
#if VERSION2
extern FTYPE            FChangeable();  /* Determine if floppy drive */
#endif
#endif

extern FTYPE            fDrivePass;     /* Executing DrivePass flag */
extern BSTYPE           bsInput;        /* Input file (library) */
WORD                    vbp;            /* Virtual Block Pointer */
char                    *vob;           /* Virtual Object Pointer */
WORD                    vcb;            /* Virtual Count of Bytes */
FTYPE                   fScrClosed;     /* Flag SCRatch file Closed */
FHTYPE                  fhScr;          /* File Handle SCRatch file */
AREATYPE                mpbparea[BPMAX];/* MaP: Blk Ptr - AREA */
PNTYPE                  mpbppv[BPMAX];  /* MaP: BlkPtr - PgVirt */
BYTE                    *mpbpovbf[BPMAXDS];
					/* MaP: BlkPtr - OffVirtByteFld */
#if CDEC20                              /* Static allocation for DEC20 */
static PAGE             rgp[BPMAXDS];   /* RanGe of Pages */
#endif                                  /* End conditional declaration */
static BPTYPE           rgbp[IBPMAX];   /* MaP: RanGe of BlkPtr */
LRUTYPE                 mpbplru[BPMAX]; /* MaP: BlkPtr - LRUnumber */
FTYPE                   rgfDirty[BPMAX];/* RanGe of Dirty Flags */
PVTYPE                  mpareapvMax[AREAMAX];
					/* MaP: AREA - PgVirtMAXimum */
PNTYPE                  rgpn[IPNMAX];   /* RanGe of PageNumbers */
IPNTYPE                 mpareaipnFirst[AREAMAX];
					/* MaP: AREA - IdxPgNumFirst */
WORD                    bpMac;          /* BlkPtrMAxCount */
WORD                    bpMacDS;        /* BlkPtrMAxCount in DataSegment */
WORD                    pnMacScr;       /* PgNumMaxCount in SCRfile */
WORD                    ipnMac;         /* IdxPgNumMaxCount */
LRUTYPE                 lruCurrent;
WORD                    saFirstVbf;     /* SegAddrFIRSTVirtByteField */
#if PERFORMANCE
WORD                    cPageFaults;    /* count of page faults */
WORD                    cMoves12;       /* count of page moves 1 to 2 */
WORD                    cMoves13;       /* count of page moves 1 to 3 */
WORD                    cMoves23;       /* count of page moves 2 to 3 */
WORD                    cMoves21;       /* count of page moves 2 to 1 */
WORD                    cMoves31;       /* count of page moves 3 to 1 */
WORD                    cMoves01;       /* count of inited pages */
FTYPE                   fPerformance;   /* TRUE -> print mem stats */
#endif
#if OSMSDOS OR OSPCDOS OR OSCPM86
static BYTE             *tempFile = "VM.TMP";
					/* VM file name */
#endif
#if OSTOPS20
static BYTE             *tempFile = "ldvmXXXXXX";
					/* Name template */
#endif
#if OSXENIX
static BYTE             *tempFile = "/tmp/ldvmXXXXXX";
					/* Name template */
#endif

    /****************************************************************
    *                                                               *
    *  Basic memory  model for  virtual memory: Virtual  addresses  *
    *  are 24 bits; an  area (upper  8  bits) and an offset within  *
    *  that area.  There is  no overlapping  from  one  area  into  *
    *  another.  The virtual  memory may be viewed as  256  65536-  *
    *  byte pieces.  The actual  implementation  uses  three areas  *
    *  for storage.  In PL/M's MEDIUM  model,  all short  pointers  *
    *  are 16 bits.  This  is a  fast and simple way of addressing  *
    *  data.  The  16-bits  are  all  addressed  within  the  data  *
    *  segment  (DGROUP).  The   three  areas   used  are:  1)  DS  *
    *  addressable 2) The rest of  available memory 3) disk space.  *
    *  The methods of moving  info around these are: 1 -> 2 1 -> 3  *
    *  3 -> 1  2 -> 1  2 -> 3.                                      *
    *                                                               *
    *  Note: Since  C  does not support long pointers to data, the  *
    *  only areas used are  DS  space and disk.  In the future,  C  *
    *  may support long pointers  to data.  Thus, the  conditional  *
    *  compilation  constant  LARGEMODEL  has  been  defined.   It  *
    *  should always be FALSE at the current time.                  *
    *                                                               *
    ****************************************************************/

    /****************************************************************
    *                                                               *
    *  InitVm:                                                      *
    *                                                               *
    *  This function takes no  arguments and returns no meaningful  *
    *  value.  It sets up virtual memory, determines the number of  *
    *  free pages in core and in the data segment, and sets up the  *
    *  page tables.                                                 *
    *                                                               *
    ****************************************************************/

InitVm()                                /* Initialize virtual memory */
  {
#if NOT CDEC20                          /* If not DEC20 C runtime */
    char                *nonVm;         /* Ptr. to non-VM heap space */
#endif                                  /* End conditional declaration */

    bpMac = 0;                          /* Assume no buffers */
#if NOT CDEC20                          /* If buffers not static */
    if(nonVm = malloc(CBNONVM))         /* If sufficient heap space */
      {
#endif                                  /* End conditional code */
	while(bpMac < BPMAXDS)          /* Loop to allocate buffers */
	  {
	    mpbparea[bpMac] = AREANIL;  /* Pages have no areas yet */
#if CDEC20                              /* Static allocation on DEC20 */
	    mpbpovbf[bpMac] = &rgp[bpMac][0];
					/* Set pointer to page */
#else                                   /* Else dynamic allocation */
	    if(!(mpbpovbf[bpMac] = malloc(CBVBF))) break;
					/* Break if out of memory */
#endif                                  /* End conditional code */
	    rgfDirty[bpMac] = FALSE;    /* Page not dirty yet */
	    mpbplru[bpMac] = 0;         /* Page has never been used */
	    bpMacDS = ++bpMac;          /* Set no. of pages in DS space */
	  }
#if NOT CDEC20                          /* If buffers not static */
	free(nonVm);                    /* Restore non-VM heap space */
      }
#endif                                  /* End conditional code */
    DEBUGVALUE(bpMac);                  /* Debug info */
    if(bpMac < CBPMIN) Fatal("Not enough memory for linker.",0x1956);
					/* Check for error */
    Setw(PNNIL,rgpn,IPNMAX);            /* Initialize disk map array */
    Setw(0,mpareapvMax,AREAMAX);        /* Initialize area size array */
    Setw(IPNNIL,mpareaipnFirst,AREAMAX);
    ipnMac = 0;
    lruCurrent = 0;                     /* Initialiaze "time" */
    vbp = 0;                            /* No current page yet */
    pnMacScr = 0;                       /* Scratch file empty */
    fScrClosed = TRUE;                  /* No scratch file yet */
    RehashRgBp();                       /* Initialize hash table */
  }

    /****************************************************************
    *                                                               *
    *  IbpLookup:                                                   *
    *                                                               *
    *  Look up virtual page and  return its physical block number.  *
    *                                                               *
    *  Input:   area    The area number being looked up             *
    *           pv      The virtual page within that area           *
    *           fCreate Create a new entry (if necessary) for this  *
    *                   in hash table                               *
    *                                                               *
    *  Return:  Index of slot in rgbp  where a block pointer lives  *
    *           corresponding to the area/pv given.                 *
    *                                                               *
    *  Note: Since the initial value of ibp will always lie in the  *
    *  range [1,510], IBPMAX is  constrained to lie in  the  range  *
    *  [255,510].  Dibp will always lie in the range [1,255].       *
    *                                                               *
    ****************************************************************/

LOCAL WORD              IbpLookup(area,pv,fCreate)
AREATYPE                area;           /* Area being looked up */
PVTYPE                  pv;             /* Particular virtual page */
FTYPE                   fCreate;        /* Create new entry flag */
  {
    register WORD       ibp;            /* Hash index */
    register WORD       bp;             /* Page number */
    WORD                dibp;           /* Hash index delta */
    long                temp;           /* Temporary */

    ENTER(IbpLookup);                   /* Debug info */
    if(!(dibp = ror(pv & ~(~0 << BYTELN),(WORD) area & 3,BYTELN))) dibp = 1;
					/* Calculate hash delta */
    ibp = dibp + rol(B2W(area),pv & 3,BYTELN);
					/* Calculate hash index */
    DEBUGVALUE(ibp);                    /* Debug info */
    DEBUGVALUE(dibp);                   /* Debug info */
    for(;;)                             /* Loop to find virtual page */
      {
	if(ibp >= IBPMAX) ibp -= IBPMAX;/* Wrap around table's end */
	bp = B2W(rgbp[ibp]);            /* Get bucket value */
	if(bp == BPNIL)                 /* If bucket empty */
	  {
	    if(fCreate)                 /* If creation requested */
	      {
		LEAVE(IbpLookup,ibp);   /* Debug info */
		return(ibp);            /* Return index of empty bucket */
	      }
	    else                        /* Else if creation not requested */
	      {
		LEAVE(IbpLookup,IBPNIL);/* Debug info */
		return(IBPNIL);         /* Page not found */
	      }
	  }
	if(mpbparea[bp] == area && mpbppv[bp] == pv)
	  {                             /* If virt. addrs match */
	    LEAVE(IbpLookup,ibp);       /* Debug info */
	    return(ibp);                /* Return bucket index */
	  }
	ibp += dibp;                    /* Try next bucket */
      }
  }

    /****************************************************************
    *                                                               *
    *  RehashRgBp:                                                  *
    *                                                               *
    *  Rehashes and builds a new page-in-core hash table.           *
    *                                                               *
    ****************************************************************/

RehashRgBp()
  {
    register WORD       bp;             /* Page number */
    register AREATYPE   area;           /* Area number */

    ENTER(RehashRgBp);                  /* Debug info */
    Setb(BPNIL,rgbp,IBPMAX);            /* Clear the table */
    for(bp = 0; bp < bpMac; ++bp)       /* Enumerate page set */
      {
	if((area = mpbparea[bp]) != AREANIL)
	  {                             /* If page is in some area */
	    rgbp[IbpLookup(area,mpbppv[bp],TRUE)] = (BPTYPE) bp;
					/* Load hash bucket with page no. */
	  }
      }
    LEAVE(RehashRgBp,0xFFFF);           /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  BpAlloc:                                                     *
    *                                                               *
    *  Allocate a free  page in  addressable  space (DS), possibly  *
    *  swapping a page to disk.                                     *
    *                                                               *
    *  Return:  Page number of free page addressable by DS.         *
    *                                                               *
    ****************************************************************/

LOCAL WORD              BpAlloc()
  {
    LRUTYPE             lru;            /* LRU queue member */
    WORD                bp;             /* Virtual page number */
    WORD                bpSave;         /* Temp. page no. storage */
#if LARGEMODEL
    POINTER             pvbf;
#endif

    ENTER(BpAlloc);                     /* Debug info */
    lru = 0xFFFF;                       /* Initialize */
    bp = 0;                             /* Initialize */
    /* "Find unused page or lru page" */
    /* "bottom of DS up.." */
    while(bp < bpMacDS)                 /* Enumerate DS pages */
      {
	if(mpbparea[bp] == AREANIL)     /* If page not used */
	  {
	    vbp = bp;                   /* This is the page we want */
	    break;                      /* Exit loop */
	  }
	if(mpbplru[bp] <= lru)          /* If page not used recently */
	  {
	    lru = mpbplru[bp];          /* Save as least recently used */
	    vbp = bp;                   /* Save page number */
	  }
	++bp;                           /* Increment page number */
      }
    /* "Now vbp is a page of DS space that will be our freed page" */
    if(mpbparea[vbp] != AREANIL)        /* If page is allocated */
      {
	lru = 0xFFFF;                   /* Initialize */
	bp = bpMacDS;                   /* Initialize */
	bpSave = vbp;                   /* Initialize */
#if LARGEMODEL
	while(bp < bpMac)               /* Enumerate non-DS pages */
	  {
	    if(mpbparea[bp] == AREANIL) /* If page not allocated */
	      {
		bpSave = bp;            /* Save virtual page number */
		break;                  /* Exit loop */
	      }
	    if(mpbplru(bp) <= lru)      /* If page less recently used */
	      {
		lru = mpbplru[bp];      /* Save lru count */
		bpSave = bp;            /* Save page number */
	      }
	    ++bp;                       /* Increment page number */
	  }
#endif
	/* bpSave is the page of a place to save vbp */
	if(rgfDirty[bpSave])            /* If page to swap out is dirty */
	  {
	    WriteBp(bpSave);            /* Write it to disk */
#if PERFORMANCE
	    ++cMoves13;                 /* Increment DS-to-disk counter */
#endif
	  }
#if LARGEMODEL
	if(vbp != bpSave)               /* If page swapped isn't page taken */
	  {
	    /* raVbf = 0; */
	    /* saVbf = saFirstVbf + SHL(DOUBLE(bpSave), 5); */
	    /* MOVW(PFromO(mpbpovbf(vbp)), pVbf, CBVBF/2); */
					/* Move from DS to non-DS */
#if PERFORMANCE
	    ++cMoves12;                 /* Increment DS-to-non-DS count */
#endif
	    mpbparea[bpSave] = mpbparea[vbp];
					/* Transfer area */
	    mpbppv[bpSave] = mpbppv[vbp];
					/* Transfer offset */
	    mpbplru[bpSave] = mpbplru[vbp];
					/* Transfer queue position */
	    rgfDirty[bpSave] = rgfDirty[vbp];
					/* Transfer dirty flag */
	  }
#endif
      }
    mpbparea[vbp] = AREANIL;            /* Assume no area */
    rgfDirty[vbp] = FALSE;              /* Assume not dirty */
    RehashRgBp();                       /* Rehash page table */
    LEAVE(BpAlloc,vbp);                 /* Debug info */
    return(vbp);                        /* Return page number */
  }

    /****************************************************************
    *                                                               *
    *  WriteBp:                                                     *
    *                                                               *
    *  Write out a dirty page.                                      *
    *                                                               *
    *  Input:   bp      Page number to write.                       *
    *                                                               *
    ****************************************************************/

LOCAL                   WriteBp(bp)
register WORD           bp;             /* Page number */
  {
    register PNTYPE     pn;             /* Scratch file page number */
    LFATYPE             lfa;            /* Logical file address */
#if OSMSDOS OR OSPCDOS OR OSCPM86
    WORD                erc;            /* Error code */
    BSTYPE              bsTmp;          /* Used to save file handle */
#endif

    ENTER(WriteBp);                     /* Debug info */
    if(fScrClosed)                      /* If scratch file closed */
      {
	fScrClosed = FALSE;             /* Open it! */
#if OSXENIX OR OSTOPS20
	tempFile = mktemp(tempFile);    /* Make temporary file name */
#endif
	close(creat(tempFile,0600));    /* Make scratch file */
#if CDEC20                              /* If DEC20 version */
	fhScr = open(tempFile,4);       /* Open it for read/write */
#else                                   /* Else if not DEC20 version */
#if CLATTICE                            /* If Lattice C version */
	fhScr = open(tempFile,0100002); /* Open it for read/write */
#else                                   /* Else if not Lattice C version */
	fhScr = open(tempFile,2);       /* Open it for read/write */
#endif                                  /* End conditional code */
#endif                                  /* End conditional code */
	if(fhScr == (FHTYPE) -1) Fatal("Cannot open temporary file",0xC0);
					/* Check for error */
#if CDEC20                              /* If DEC20 version */
	_fotab[jfnof(fhScr)]._flg |= RAW;
					/* Set file to raw mode */
#endif                                  /* End DEC20 code */
#if DEBUG
	fprintf(stderr,"%s has been created\n",tempFile);
					/* Scratch file name */
#endif
#if OSMSDOS OR OSPCDOS OR OSCPM86
	erc = (WORD) SCPCurrentDisk() + 'A';
					/* Get drive letter */
#if VERSION2
	if(FChangeable((BYTE) erc))     /* If default drive is a floppy */
	  {
#endif
	    fprintf(stderr,"%s has been created.",tempFile);
					/* Scratch file name */
	    NEWLINE(stderr);            /* Newline */
	    fprintf(stderr,"Do not change diskette in drive %c:",(BYTE) erc);
					/* Message */
	    NEWLINE(stderr);            /* Newline */
	    fflush(stderr);             /* Force message to screen */
#if VERSION2
	  }
#endif
#endif
      }
    pn = rgpn[mpareaipnFirst[B2W(mpbparea[bp])] + mpbppv[bp]];

    if(pn == PNNIL)
      {
	pn = pnMacScr++;
	rgpn[mpareaipnFirst[B2W(mpbparea[bp])] + mpbppv[bp]] = pn;
      }
    lfa = (long) pn << SHPLFALOW;       /* Calculate lfa */
    if(lfa != lseek(fhScr,lfa,0)) Fatal("scratch write err",0xC1);
    if(write(fhScr,mpbpovbf[bp],CBVBF) != CBVBF)
      {                                 /* If write fails */
	fDrivePass = FALSE;             /* Clear DrivePass flag */
	Fatal("Out of space on scratch file",0xC1);
					/* Fatal error */
      }
    rgfDirty[bp] = FALSE;               /* Clear flag */
    LEAVE(WriteBp,0xFFFF);              /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  ObGetPn:                                                     *
    *                                                               *
    *  Return  in-core  page  number  of area/pv.  May necessitate  *
    *  swapping.                                                    *
    *                                                               *
    *  Input:   area    The area from which to retrieve.            *
    *           pv      Virtual page within area to retrieve.       *
    *           fDirty  Flag whether this  will be dirtied (if you  *
    *                   know ahead of time).                        *
    *                                                               *
    *  Return:  Page  number (in-core) of  corresponding  area/pv.  *
    *           This page number is addressable via DS.             *
    *                                                               *
    ****************************************************************/

char                    *ObGetPn(area,pv,fDirty)
AREATYPE                area;           /* Area from which to retrieve */
PVTYPE                  pv;             /* Virtual page within area */
FTYPE                   fDirty;         /* Dirty page flag */
  {
    LRUTYPE             dlru;
    IBPTYPE             ibp;
    IPNTYPE             ipnFirst;
    register WORD       bp;             /* Page number */
    PNTYPE              pn;             /* Scratch file page number */
    LFATYPE             lfa;            /* Logical file address */
#if LARGEMODEL
    POINTER             pvbf;
#endif
#if DEBUG
    BSTYPE              bsTmp;          /* File handle */
#endif

    ENTER(ObGetPn);                     /* Debug info */
    ASSERT(area != 0);                  /* Cannot use AREANIL */
    if(mpbparea[vbp] == area && mpbppv[vbp] == pv)
      {                                 /* If asked for current page */
	if(fDirty) rgfDirty[vbp] = TRUE;/* Set dirty flag if asked to */
	LEAVE(ObGetPn,mpbpovbf[vbp]);   /* Debug info */
	return(mpbpovbf[vbp]);          /* Return pointer */
      }
    ibp = IbpLookup(area, pv, FALSE);   /* Look up requested page */
    if(ibp != IBPNIL)                   /* If requested page is in memory */
      {
	bp = B2W(rgbp[ibp]);            /* Get the page number */
#if LARGEMODEL
	if(bp >= bpMacDS)               /* If page not in DS space */
	  {
	    mpbplru[bp] = lruCurrent;   /* Set LRU rating for page */
	    vbp = B2W(BpAlloc());       /* Allocate a page in DS space */
	    /* Must have code to retrieve from non-DS memory */
#if PERFORMANCE
	    ++cMoves21;                 /* Increment non-DS-to-DS count */
#endif
	    ibp = IbpLookup(area, pv, FALSE);
					/* Look up address in hash table */
	    rgbp[ibp] = (BPTYPE) vbp;   /* Save page number in table */
	    mpbparea[vbp] = area;       /* Save page area in table */
	    mpbppv[vbp] = pv;           /* Save page no. within area */
	    rgfDirty[vbp] = rgfDirty[bp];
					/* Copy dirty flag */
	    mpbparea[bp] = AREANIL;     /* Clear old entry */
	    rgfDirty[bp] = FALSE;
	  }
	else                            /* Else if page in DS space */
#endif
	  vbp = bp;                     /* Save page number */
      }
    else                                /* Else if page is not in memory */
      {
#if PERFORMANCE
	++cPageFaults;                  /* Increment page fault count */
#endif
	vbp = BpAlloc();                /* Allocate a page in DS space */
	ibp = IbpLookup(area,pv,TRUE);  /* Make new entry in hash table */
	DEBUGMSG("Page not found: ");   /* Message */
	DEBUGVALUE(ibp);                /* Debug info */
	rgbp[ibp] = (BPTYPE) vbp;       /* Save page number */
	mpbparea[vbp] = area;           /* Save area */
	mpbppv[vbp] = pv;               /* Save page number within area */
#if LIBMSDOS
	if(B2W(area) < AREAFLB)         /* If page is in temp file */
	  {
#endif
	    if(pv >= mpareapvMax[B2W(area)])
	      {                         /* If page no. above top bound */
		DEBUGVALUE(area);       /* Debug info */
		DEBUGVALUE(pv);         /* Debug info */
		/* NOTE:  The following error message is sometimes bogus.
		*  For example, if the user exceeds the maximum number of
		*  32-bit fixups the linker is able to perform, this is
		*  the error message he will receive.
		*/
		Fatal("Attempt to access data outside of segment bounds.",
		  0xC2);                /* Die */
	      }
	    ipnFirst = mpareaipnFirst[area];
					/* Get ipn of 1st page in area */
	    ASSERT(ipnFirst != IPNNIL); /* Should never happen */
	    pn = rgpn[ipnFirst + pv];   /* Get disk page no. of desired page */
	    if(pn == PNNIL)             /* If unallocated page */
	      {
		Setb('\0',PFromO(mpbpovbf[vbp]), CBVBF);
					/* Clear the page */
#if PERFORMANCE
		++cMoves01;             /* Increment allocation count */
#endif
	      }
	    else                        /* Else if allocated page */
	      {
		lfa = (LFATYPE) pn << SHPLFALOW;
					/* Calculate lfa */
		if(lfa != lseek(fhScr,lfa,0)) Fatal("scratch read err",0xC4);
		if(read(fhScr,PFromO(mpbpovbf[vbp]),CBVBF) != CBVBF)
		  {                     /* If read fails */
		    Fatal("Unexpected end-of-file on scratch file",0xC4);
					/* Fatal error */
		  }
#if PERFORMANCE
		++cMoves31;             /* Increment disk-to-DS count */
#endif
	      }
#if LIBMSDOS
	  }
	else
	  {
	    fseek(bsInput,(LFATYPE) pv << SHPLFALOW,0);
					/* Seek to desired offset */
	    fread(PFromO(mpbpovbf[vbp]),1,CBVBF,bsInput);
					/* Read page in */
	    if(ferror(bsInput)) Fatal("Unexpected EOF on library",0xC5);
					/* Check for error */
	  }
#endif
      }
    ++lruCurrent;                       /* Increment LRU number */
    if(!lruCurrent)                     /* If counter overflow */
      {
	dlru = 0xFFFF;                  /* Set dlru to 64K - 1 */
	for(bp = 0; bp < bpMac; ++bp)   /* Loop to find smallest lru value */
	  {
	    if(mpbparea[bp] != AREANIL) /* If page is allocated */
	      {
		if(mpbplru[bp] < dlru) dlru = mpbplru[bp];
					/* If lru value smaller, save it */
	      }
	  }
	if(dlru < 0x8000) dlru = 0x8000;/* Clamp at 32K */
	for(bp = 0; bp < bpMac; ++bp)   /* Loop to readjust lru values */
	  {
	    if(mpbparea[bp] != AREANIL) /* If page is allocated */
	      {
		if(mpbplru[bp] < dlru) mpbplru[bp] = 0;
					/* Zero if less than dlru */
		else mpbplru[bp] -= dlru;
					/* Otherwise subtract dlru */
		if(mpbplru[bp] >= lruCurrent) lruCurrent = mpbplru[bp] + 1;
					/* Make lruCurrent largest val + 1 */
	      }
	  }
      }
    mpbplru[vbp] = lruCurrent;          /* This page is most recently used */
    if(fDirty) rgfDirty[vbp] = TRUE;    /* Set dirty flag if asked to */
    LEAVE(ObGetPn,mpbpovbf[vbp]);       /* Debug info */
    return(mpbpovbf[vbp]);              /* Return pointer to page */
  }

    /****************************************************************
    *                                                               *
    *  ObGetVa:                                                     *
    *                                                               *
    *  Get  physical address  corresponding to  area/pv.  This may  *
    *  necessitate  swapping.  The returned address  is within the  *
    *  current data segment.                                        *
    *                                                               *
    *  Input:   area    Number of area from which to retrieve.      *
    *           va      Virtual address within area.                *
    *           fDirty  Flag  specifying whether or  not the whole  *
    *                   page is to be dirtied.                      *
    *                                                               *
    *  Return:  Address within DS corresponding to area/va.         *
    *                                                               *
    ****************************************************************/

char                    *ObGetVa(area,va,fDirty)
AREATYPE                area;           /* Area from which to retrieve */
VATYPE                  va;             /* Virtual address within that area */
FTYPE                   fDirty;         /* Specify if page is to be dirtied */
  {
    return(ObGetPn(area,va >> SHBTOPN,fDirty) + (va & MASKDB));
  }

    /****************************************************************
    *                                                               *
    *  ReadPvs:                                                     *
    *                                                               *
    *  Read consecutive pages from virtual area/pv into object ob.  *
    *                                                               *
    *  Input:   ob      Destination address of pages.               *
    *           area    Area number of source pages.                *
    *           pv      Virtual page  within  area of beginning of  *
    *                   source.                                     *
    *           cpv     Number of pages to transfer.                *
    *                                                               *
    ****************************************************************/

ReadPvs(ob,area,pv,cpv)
char                    *ob;            /* Destination */
AREATYPE                area;           /* Source area */
PVTYPE                  pv;             /* Source area starting page */
PVTYPE                  cpv;            /* Count of pages to read */
  {
    PVTYPE              cpvChunk;
    LFATYPE             lfa;            /* Logical file address */
    IBPTYPE             ibp;            /* Page hash table index */
    PNTYPE              pn;             /* Scratch file page number */
    IPNTYPE             ipnFirst;       /* Index to first page in area */
    POINTER             pVbf;

    ENTER(ReadPvs);                     /* Debug info */
    ipnFirst = mpareaipnFirst[area];    /* Get index to first page in area */
    if(ipnFirst == IPNNIL) ErrHlt(0xC6);/* Should never happen */
    while(cpv)                          /* While there are pages to read */
      {
	ibp = IbpLookup(area,pv,FALSE); /* Look up page in hash table */
	if(ibp != IBPNIL)               /* If entry exists (page in memory) */
	  {
	    pVbf = mpbpovbf[B2W(rgbp[ibp])];
					/* Calc memory address of source */
	    Move(CBVBF,pVbf,PFromO(ob));/* Move from source to dest 1 page */
	    ob += CBVBF;                /* Increment dest pointer */
	    ++pv;                       /* Increment page number */
	    --cpv;                      /* Decrement page count */
	  }
	else                            /* Else if page not in memory */
	  {
	    if(pv > mpareapvMax[area]) ErrHlt(0xC7);
					/* Should never happen */
	    pn = rgpn[ipnFirst + pv];   /* Get disk page number */
	    if(pn == PNNIL)             /* If page unallocated before */
	      {
		Setb('\0',PFromO(ob),CBVBF);
					/* Zero the destination */
		++pv;                   /* Increment the page number */
		--cpv;                  /* Decrement the page count */
		ob += CBVBF;            /* Increment the dest pointer */
	      }
	    else                        /* Else if page allocated before */
	      {
		lfa = (LFATYPE) pn << SHPLFALOW;
					/* Calculate lfa */
		cpvChunk = 1;
		--cpv;                  /* Decrement page count */
		++pv;                   /* Increment page number */
		while(cpv && rgpn[ipnFirst + pv] == pn + 1)
		  {                     /* While pages are consecutive */
		    if(IbpLookup(area,pv,FALSE) != IBPNIL) break;
					/* If next page in memory, break */
		    ++cpvChunk;         /* Increment chunk size */
		    --cpv;              /* Decrement page count */
		    ++pv;               /* Increment page number */
		    ++pn;               /* Increment file page number */
		  }
		if(lfa != lseek(fhScr,lfa, 0)) Fatal("scratch read err",0xC8);
					/* Seek to offset in scratch file */
		if(read(fhScr,PFromO(ob),(WORD) cpvChunk*CBVBF)
		  != cpvChunk*CBVBF) ErrHlt(0xC8);
					/* Very bad if read fails */
		ob += cpvChunk*CBVBF;   /* Increment dest. pointer */
	      }
	  }
      }
    LEAVE(ReadPvs,0xFFFF);              /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  MapSpan:                                                     *
    *                                                               *
    *  "Set public  vcb to min (number of  bytes left in page from  *
    *  va, cb).  Also forces  that  page  to  be in DS.  vob is DS  *
    *  pointer to free space."                                      *
    *                                                               *
    *  Input:   area    Area number of virtual address.             *
    *           va      Virtual address within area.                *
    *           cb      Count of bytes that we need to move.        *
    *           fDirty  "Whether we should dirty the page."         *
    *                                                               *
    ****************************************************************/

MapSpan(area,va,cb,fDirty)
AREATYPE                area;           /* Virtual area */
VATYPE                  va;             /* Offset within that area */
WORD                    cb;             /* Count of bytes */
FTYPE                   fDirty;         /* Dirty flag */
  {
    ENTER(MapSpan);                     /* Debug info */
    vob = ObGetVa(area,va,fDirty);      /* Fetch from VM */
    vcb = CBVBF - (va & MASKDB);        /* Set size of object */
    if(vcb > cb) vcb = cb;              /* Clamp size to max of cb */
    LEAVE(MapSpan,0xFFFF);              /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  MoveToVm:                                                    *
    *                                                               *
    *  Move a piece of data into a virtual memory area/va.          *
    *                                                               *
    *  Input:   cb      Count of bytes to be moved.                 *
    *           obData  Address of data to be moved.                *
    *           area    Area number where data is to be placed.     *
    *           va      Virtual address within area of destination. *
    *                                                               *
    ****************************************************************/

MoveToVm(cb,obData,area,va)
WORD                    cb;             /* Count of bytes to move */
char                    *obData;        /* Address of bytes to be moved */
AREATYPE                area;           /* Area to move to */
VATYPE                  va;             /* Offset to move to */
  {
    ENTER(MoveToVm);                    /* Debug info */
    while(cb)                           /* While there are bytes to move */
      {
	MapSpan(area,va,cb,TRUE);       /* Ask for cb bytes of area in DS */
	cb -= vcb;                      /* Subtract actual number */
	va += vcb;                      /* Increment offset appropriately */
	Move(vcb,obData,vob);           /* Move data to VM in memory */
	obData += vcb;                  /* Increment source pointer */
      }
    LEAVE(MoveToVm,0xFFFF);             /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  InitArea:                                                    *
    *                                                               *
    *  Set up an area of virtual memory having a maximum number of  *
    *  pages.                                                       *
    *                                                               *
    *  Input:   area    Area number to be allocated.                *
    *           pvMax   Maximum number of virtual  pages that this  *
    *                   area can grow to.                           *
    *                                                               *
    ****************************************************************/

InitArea(area,pvMax)
AREATYPE                area;           /* Area number to be allocated */
PVTYPE                  pvMax;          /* Maximum number of pages in area */
  {
    register WORD       a;              /* Area */

    ENTER(InitArea);                    /* Debug info */
    a = B2W(area);                      /* Convert to WORD */
    if(!a || a >= AREAFLB || ipnMac + pvMax > IPNMAX)
      Fatal("Program size exceeds capacity of linker",0xC9);
					/* Check if not enough space */
    mpareapvMax[a] = pvMax;             /* Set maximum number of pages */
    mpareaipnFirst[a] = ipnMac;         /* Set starting page */
    ipnMac += pvMax;                    /* Increment count of pages total */
    LEAVE(InitArea,0xFFFF);             /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  FinishVm:                                                    *
    *                                                               *
    *  Close and delete vitrual memory scratch file.                *
    *                                                               *
    ****************************************************************/

FinishVm()
  {
    WORD                erc;            /* Error code */

    ENTER(FinishVm);                    /* Debug info */
    if(!fScrClosed)                     /* If scratch file open */
      {
	erc = close(fhScr);             /* Close it */
	erc |= unlink(tempFile);        /* Delete it */
	if(erc) ErrHlt(0xCA);           /* Should never happen */
	fScrClosed = TRUE;              /* Scratch file closed */
      }
#if PERFORMANCE
    if(fPerformance)
      {
	NEWLINE(stderr);
	fputs("Virtual Memory statistics:",stderr);
	NEWLINE(stderr);
	fprintf(stderr,"Page faults:           %d",cPageFaults);
	NEWLINE(stderr);
	fprintf(stderr,"# pages allocated:     %d",cMoves01);
	NEWLINE(stderr);
#if LARGEMODEL
	fprintf(stderr,"# pages DS to Memory:  %d",cMoves12);
	NEWLINE(stderr);
	fprintf(stderr,"# pages Memory to DS:  %d",cMoves21);
	NEWLINE(stderr);
	fprintf(stderr,"# pages Mem. to Disk:  %d",cMoves23);
	NEWLINE(stderr);
#endif
	fprintf(stderr,"# pages DS to Disk:    %d",cMoves13);
	NEWLINE(stderr);
	fprintf(stderr,"# pages Disk to DS:    %d",cMoves31);
	NEWLINE(stderr);
	fflush(stderr);
      }
#endif
    LEAVE(FinishVm,0xFFFF);             /* Debug info */
  }
