    /****************************************************************
    *                                                               *
    *                           XENLIB.C                            *
    *                                                               *
    ****************************************************************/

#include                "minlit.h"      /* Types and constants */
#include                "bndtrn.h"      /* More types and constants */
#include                "lnkio.h"       /* I/O definitions */

#if CDEC20                              /* Only 6 sig. char.s */
#define vrpropNewList   l00001          /* Redefine local name */
#endif                                  /* End DEC20 code */

extern                  ErrHlt();       /* Fatal error procedure */
extern                  FatalFile();    /* Fatal file */
extern                  SaveInput();    /* Save input file name and address */
extern PROPTYPE         PropSymLookup();/* Symbol table lookup */
extern                  DrivePass();    /* Pass 1 and 2 driver function */
extern                  ProcP1();       /* Pass 1 control function */
extern                  EnSyms();       /* Enumerate symbols */
extern BYTE             *ObGetVa();     /* Fetch from virtual memory */
extern BYTE             *ObGetPn();     /* Fetch virtual page */
extern WORD             WGets();        /* Get a word of input */
extern BYTE             Gets();         /* Get a byte of input */

extern BSTYPE           bsLst;          /* List file descriptor */
extern BSTYPE           bsInput;        /* Input file descriptor */
extern BYTE             *vob;           /* Set by MapSpan() */
extern RBTYPE           vrhte;          /* Virtual hash table address */
extern RBTYPE           vrhteFile;      /* Virtual hash table addr for file */
extern RBTYPE           vrpropTailFile; /* Virt. addr. of end of file list */
extern RBTYPE           rprop1stFile;   /* First file in list */
extern RBTYPE           vrpropFile;     /* Virt. addr. of current file */
extern FTYPE            fDrivePass;     /* True if(executing DrivePass() */
extern FHTYPE           ifhLast;        /* Last file index */
extern FTYPE            fLoWordFirst;   /* Low order word first in ranlib */

static RBTYPE           vrpropNewList;  /* New file list */
WORD                    ifhLibMac;      /* Max. library index value */
RBTYPE                  mpifhrhte[IFHLIBMAX];
					/* Map lib index to hash tab addr */
static WORD             mpifhcSyms[IFHLIBMAX];
					/* Map lib index to no. of pub syms */
FTYPE                   vfUndefsSeen;   /* True if(undefined symbols seen */
FTYPE                   vfExtractedAFile;
					/* True if(a file extracted */
FHTYPE                  ifhLibCur;      /* Index of current library */
FTYPE                   vfLibrarySearch;/* True if(searching libraries */
FTYPE                   fLibPass;       /* True if(input is from library */

static FTYPE            vfLibOpen;      /* True if(library open */
static FTYPE            fSkipLib;       /* True if(skipping library */

    /****************************************************************
    *                                                               *
    *  The following comment is a quote from Mark Zbikowski's PL/M  *
    *  version of the linker.  I pass it along without comment.     *
    *                                                               *
    *  "XENIX libraries  are  dumb  --  they  have  the symbols in  *
    *  a  particular  area  in  no  particular  order.   The  most  *
    *  efficient way  to  search  these  symbols  is (ick) to read  *
    *  the symbols, one by one, and test to  see  if  that  symbol  *
    *  has attrUndef...  gross, nyet?"                              *
    *                                                               *
    ****************************************************************/

LOCAL                   LibErr(pb,cb,pfErr)
BYTE                    *pb;            /* File name */
WORD                    cb;             /* Name length */
FTYPE                   *pfErr;         /* Pointer to error return */
  {
    SDTYPE              sd;             /* File name struct */

    if(pfErr) *pfErr = TRUE;            /* Error trapped */
    else                                /* Else if not trapping errors */
      {
	sd.pb = pb;                     /* Set name pointer */
	sd.cb = cb;                     /* Set name length */
	FatalFile(&sd,0x30);            /* Fatal error */
      }
  }

    /****************************************************************
    *                                                               *
    *  OpenLibrary:                                                 *
    *                                                               *
    *  This function opens a library archive for input.             *
    *                                                               *
    ****************************************************************/

OpenLibrary(pb,cb,ifhLib,pfErr)
BYTE                    *pb;            /* Pointer to file name */
WORD                    cb;             /* Length of file name */
FHTYPE                  ifhLib;         /* Library index */
FTYPE                   *pfErr;         /* Pointer to error flag */
  {
    LFATYPE             lfa;            /* Logical file address */
    SBTYPE              sbFile;         /* File name */

    ENTER(OpenLibrary);                 /* Debug info */
    if(pfErr) *pfErr = FALSE;           /* Assume no error */
    Move(cb,pb,sbFile);                 /* Copy file name */
    sbFile[cb] = '\0';                  /* Null-terminate it */
    if((bsInput = fopen(sbFile,RDBIN)) == NULL)
      {                                 /* If error opening file */
	LibErr(pb,cb,pfErr);            /* Process error */
	return;                         /* Leave */
      }
    SETRAW(bsInput);                    /* Binary file */
    if(WGets() != MAGICARCHIVE)         /* If first word isn't magic number */
      {
#if DEBUG
	fprintf(stderr,"** BAD MAGIC NUMBER **\n");
#endif
	fclose(bsInput);                /* Close file */
	LibErr(pb,cb,pfErr);            /* Process error */
	return;                         /* Leave */
      }
    GetBytes(&sbFile[1],14);            /* Read name field */
    sbFile[0] = '\011';                 /* Set length */
#if DEBUG
    OutSb(stderr,sbFile);               /* Name */
    NEWLINE(stderr);                    /* Newline */
    fflush(stderr);                     /* Force message out */
#endif
    if(SbCompare(sbFile,"\011__.SYMDEF",0))
      {                                 /* If no match */
	fclose(bsInput);                /* Release file handle */
	LibErr(pb,cb,pfErr);            /* Process error */
	return;                         /* Return */
      }
    lfa = (LFATYPE) ARDICTLEN;          /* Set file address */
    fseek(bsInput,lfa,0);               /* Seek in file */
    mpifhcSyms[ifhLib] = WGets();       /* Get symbol count */
    DEBUGVALUE(mpifhcSyms[ifhLib]);     /* Debug info */
    LEAVE(OpenLibrary,0xFFFF);          /* Debug info */
  }

    /****************************************************************
    *                                                               *
    *  LibrarySearch:                                               *
    *                                                               *
    ****************************************************************/

LibrarySearch()
  {
    WORD                erc;            /* Error code */
    RBTYPE              vrpTmp1stFile;  /* First file in list (temp) */
    extern              FindAnUndef();  /* Library lookup, maybe? */

    vfUndefsSeen = TRUE;                /* Assume there are undefined syms */
    vfLibrarySearch = TRUE;             /* We are searching libraries */
    vfLibOpen = FALSE;                  /* No library open yet */
    ifhLibCur = 0;                      /* Initialize index */
    while(ifhLibCur < ifhLibMac && vfUndefsSeen)
      {                                 /* While libraries and undefs remain */
	while(vfUndefsSeen)             /* While there are undefined symbols */
	  {
	    vfExtractedAFile = FALSE;   /* Assume no extractions */
	    vfUndefsSeen = FALSE;       /* Assume no more undefs */
	    fSkipLib = FALSE;           /* Do not skip the library */
	    EnSyms(FindAnUndef,ATTRUND);/* Enumerate undefined symbols */
	    if(vfUndefsSeen) ScanLibrary();
					/* Scan the library */
	    if(!vfExtractedAFile) break;/* Break if no files extracted */
	    ifhLast = ifhLibCur;        /* Save current index value */
	    vrpTmp1stFile = rprop1stFile;
					/* Save pointer to first file */
	    rprop1stFile = vrpropNewList;
					/* Get pointer to new list */
	    fLibPass = TRUE;            /* Passing over a library */
	    DrivePass(ProcP1);          /* Do pass 1 on library */
	    fLibPass = FALSE;           /* No longer in library */
	    rprop1stFile = vrpTmp1stFile;
					/* Restore pointer */
	  }
	if(vfLibOpen)                   /* If library open */
	  {
	    fclose(bsInput);            /* Close file */
	  }
	vfLibOpen = FALSE;              /* No library open */
	++ifhLibCur;                    /* Increment index */
      }
    vfLibrarySearch = FALSE;            /* Not searching libraries */
  }

    /****************************************************************
    *                                                               *
    *  FindAnUndef:                                                 *
    *                                                               *
    ****************************************************************/

FindAnUndef(hte,prop,rhte,rprop,fNewHte)
HTETYPE                 hte;            /* Hash table entry */
PROPTYPE                prop;           /* Property list entry */
RBTYPE                  rhte;           /* Virtual addr of hte */
RBTYPE                  rprop;          /* Virtual addr of prop list entry */
FTYPE                   fNewHte;        /* True if 1st time this hte */
  {
    if(!fSkipLib) vfUndefsSeen = TRUE;  /* Set flag if not skipping library */
  }

    /****************************************************************
    *                                                               *
    *  ScanLibrary:                                                 *
    *                                                               *
    ****************************************************************/

ScanLibrary()
  {
    AHTETYPE            *ahte;          /* Pointer to hash table entry */
    LFATYPE             lfa;            /* Logical file address */
    SBTYPE              sb;             /* Length-prefixed string */
    AHTETYPE            *ahteLib;       /* Library hash table entry */
    FTYPE               fErr;           /* Error flag */
    WORD                i;              /* Index variable */
    APROPUNDEFTYPE      *apropUndef;    /* Pointer to property cell */

    if(!vfLibOpen)                      /* If no library open */
      {
	ahteLib = (AHTETYPE *) ObGetVa(AREASYMS,mpifhrhte[ifhLibCur],TRUE);
					/* Fetch hash table entry */
	OpenLibrary(&ahteLib->cch[1],B2W(ahteLib->cch[0]),ifhLibCur,&fErr);
					/* Open library */
	if(fErr)                        /* If error opening library */
	  {
	    fprintf(stderr,"Error accessing library ");
					/* Message */
	    OutSb(stderr,ahteLib->cch); /* Library name */
	    NEWLINE(stderr);            /* Newline */
	    fflush(stderr);             /* Force message to screen */
	    fSkipLib = TRUE;            /* Skip this library */
	    return;                     /* Return */
	  }
	vfLibOpen = TRUE;               /* Library open */
      }
    /* "now, read in the lib symbols, one by one..." */
    lfa = ARDICTLOC;                    /* Set file address */
    fseek(bsInput,lfa,0);               /* Seek to address */
    for(i = 0; i < mpifhcSyms[ifhLibCur]; ++i)
      {                                 /* Loop to read symbols */
	sb[0] = Gets();                 /* Read length byte */
	GetBytes(&sb[1],B2W(sb[0]));    /* Read rest of symbol */
#if DEBUG
	fprintf(stderr,"Symbol \"");    /* Message */
	OutSb(stderr,sb);               /* Symbol */
	OutByte(stderr,'"');            /* Close quotes */
	NEWLINE(stderr);                /* Newline */
	fflush(stderr);                 /* Force message out */
#endif
	if(fLoWordFirst)                /* If low order word first */
	  {
	    lfa = (LFATYPE) WGets();    /* Get low-order word of file addr */
	    lfa += (LFATYPE) WGets() << WORDLN;
					/* Get high-order word of file addr */
	  }
	else                            /* Else if high order word first */
	  {
	    lfa = (LFATYPE) WGets() << WORDLN;
					/* Get high-order word of file addr */
	    lfa += (LFATYPE) WGets();   /* Get low-order word of file addr */
	  }
	if((apropUndef = PropSymLookup(sb,ATTRUND,FALSE)) != PROPNIL)
	  {
	    if(!apropUndef->au_len)     /* If not an undefined data item */
	      {
		ahte = ObGetVa(AREASYMS,mpifhrhte[ifhLibCur],FALSE);
					/* Fetch from vm */
		lfa += (LFATYPE) ARHEADLEN;
					/* Adjust file address for header */
		DEBUGVALUE(lfa);        /* Debug info */
		SaveInput(ahte->cch,lfa,ifhLibCur);
					/* Save file name and location */
		if(!vfExtractedAFile) vrpropNewList = vrpropTailFile;
		vfExtractedAFile = TRUE;/* A file has been extracted */
	      }
	  }
      }
  }
